"""
GeoJSON annotations are aimed to be compatible with QuPath.
"""

from __future__ import annotations
import math
from typing import Type
from enum import Enum
import numpy as np
import json

import shapely
from shapely.geometry import shape, Polygon
from mmm.typing_utils import get_colors


class AnnotationType(Enum):
    # Perfect for a detection task
    DetectionBox = "DetectionBox"
    # Perfect for a semantic segmentation task with close boundaries
    Contour = "Contour"
    # Single object with multiple regions
    MultiContour = "MultiContour"


class GeoAnno:
    """
    Wraps a single QuPath geojson annotation such as

    ```
    {
        "type": "Feature",
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [
                        7624,
                        9051
                    ],
                    [
                        7624,
                        9227
                    ],
                    [
                        7845,
                        9227
                    ],
                    [
                        7845,
                        9051
                    ],
                    [
                        7624,
                        9051
                    ]
                ]
            ]
        },
        "properties": {
            "object_type": "annotation",
            "classification": {
                "name": "Dust_8.1.1",
                "colorRGB": -4987213
            },
            "isLocked": false
        }
    }
    ```
    """

    @classmethod
    def from_shapely(cls, shape: shapely.Geometry, class_info: tuple[str, int, int] | None = None) -> GeoAnno:
        """
        class_info is given as (class_name, max_classes, class_index)
        """
        json_dict = {
            "type": "Feature",
            "geometry": shapely.geometry.mapping(shape),
            "properties": {
                "object_type": "annotation",
            },
        }
        if class_info:
            class_color = get_colors(class_info[1])[class_info[2]]  # between 0 and 1
            json_dict["properties"]["classification"] = {
                "name": class_info[0],
                "color": [int(x * 255) for x in class_color],
            }
        return cls(json_dict)

    @classmethod
    def rectangle_builder(
        cls: Type[GeoAnno],
        left_upper: tuple[int, int],
        right_lower: tuple[int, int],
        class_name="defaultclass",
    ) -> GeoAnno:
        return cls(
            {
                "type": "Feature",
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [left_upper[0], left_upper[1]],
                            [right_lower[0], left_upper[1]],
                            [right_lower[0], right_lower[1]],
                            [left_upper[0], right_lower[1]],
                            [left_upper[0], left_upper[1]],
                        ]
                    ],
                },
                "properties": {
                    "object_type": "annotation",
                    "classification": {"name": class_name, "color": [200, 0, 0]},
                },
            }
        )

    def __init__(self, json_dict: dict, origin_id: str | None = None) -> None:
        self.wsi_id: str | None = origin_id
        self.shape = shape(json_dict["geometry"])

        self.properties: dict = json_dict["properties"] if "properties" in json_dict else {}

        # QuPath needs a color
        try:
            self.color: list[int] = self.properties["classification"]["color"]
            assert len(self.color) == 3
        except KeyError:
            self.color = [200, 0, 0]

        if self.shape.geom_type == "LineString":
            # Convert to Polygon, because a LineString has no area and we depend on that
            self.shape = self.shape.buffer(np.sqrt(self.shape.length))

    def set_class_name(self, class_name: str):
        if "classification" not in self.properties:
            self.properties["classification"] = {
                "name": class_name.lower(),
                "color": self.color,
            }
        else:
            self.properties["classification"]["name"] = class_name.lower()

    def get_class_name(self) -> str | None:
        try:
            return self.properties["classification"]["name"].lower()
        except KeyError:
            return None

    def get_enclosing_region(self) -> tuple[tuple[int, int], tuple[int, int]]:
        bounds: tuple[int, int, int, int] = self.shape.bounds  # type: ignore
        return (int(bounds[0]), int(bounds[1])), (
            int(bounds[2] - bounds[0]),
            int(bounds[3] - bounds[1]),
        )

    def overlaps_with(self, other_shape: Polygon, overlap_threshold: float = 0.0) -> bool:
        if self.shape.intersects(other_shape):
            intersection_shape = self.shape.intersection(other_shape)
            return intersection_shape.area > (self.shape.area * overlap_threshold)
        return False

    def get_annotation_type(self) -> AnnotationType:
        """
        If the annotation looks like a perfect box, it returns AnnotationType.DetectionBox.
        Otherwise: AnnotationType.Contour or AnnotationType.MultiContour
        """
        if math.isclose(self.shape.area, self.shape.minimum_rotated_rectangle.area):
            return AnnotationType.DetectionBox
        elif self.shape.geom_type == "Polygon":
            return AnnotationType.Contour
        elif self.shape.geom_type == "MultiPolygon":
            return AnnotationType.MultiContour
        else:
            raise NotImplementedError(f"Not implemented for: {self.shape.geom_type}")

    def visualize_as_pyplot(self):
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        if c := self.get_class_name():
            ax.set_title(f"Class {c}")
        ax.plot(*self.shape.exterior.xy)
        return fig

    def to_geojson(self) -> dict:
        res = {
            "type": "Feature",
            "geometry": shapely.geometry.mapping(self.shape),
        }
        if self.properties:
            res["properties"] = self.properties
        return res

    def __repr__(self) -> str:
        return f"GeoAnno:\n{self.shape.geom_type=};{self.shape.area=}\n{json.dumps(self.properties, indent=2)}"



"""
Frequently required training utilities
"""

from mmm.trainer.MTLTrainer import MTLTrainer, TaskPurpose, DataSplit
from mmm.task_sampling import CyclicTaskSampler, BalancedTaskSampler
from mmm.trainer.Loop import MultistepMode, LinearMultistep, FixedMultistep



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from torchvision.utils import make_grid

from typing import Literal

from .SharedBlock import SharedBlock


def make_grid_for_supercase(training_ims, supercase_indices, group_index, grouper_weights):
    vis_indices = torch.where(supercase_indices == group_index)[0].cpu()
    rows = int(np.sqrt(len(vis_indices)))
    # cols = training_ims[vis_indices].shape[0] // rows
    grid_img = make_grid(training_ims[vis_indices], nrow=rows)

    vis_cases_weights = grouper_weights[vis_indices]
    # with multiple weights available for one image each row is a list of the corresponding weights
    weight_chunks = []
    for vis_case in vis_cases_weights:
        weight_chunks.append(",".join([f"{single_w:.2f}" for single_w in vis_case]))
    weight_rows = [f"[{w}]" for w in weight_chunks]
    weight_rows[-1] = weight_rows[-1] + " " * (len(weight_rows[0]) - len(weight_rows[-1]))
    weight_str = "\n".join(weight_rows)
    return grid_img, weight_str, vis_indices


class AttentionPoolingReducer(nn.Module):
    """
    Attention pooling akin to ABMIL https://arxiv.org/abs/1802.04712
    an be used with gated attention (default) or regular attention.
    Will apply a linear layer after the attention pooling to ensure embedding_dim is met.
    """

    def __init__(self, embedding_dim, gated: bool = True, num_heads: int = 1) -> None:
        super(AttentionPoolingReducer, self).__init__()
        self.embedding_dim = embedding_dim

        # Different number of attention heads
        self.num_heads = num_heads

        # boolean if to use gated attention or not
        self.gated = gated
        if gated:
            self.attention_v = nn.Sequential(nn.Linear(self.embedding_dim, self.embedding_dim // 3), nn.Tanh())
            self.attention_u = nn.Sequential(nn.Linear(self.embedding_dim, self.embedding_dim // 3), nn.Sigmoid())
            self.attention = nn.Linear(self.embedding_dim // 3, num_heads)
        else:
            self.attention = nn.Sequential(
                nn.Linear(self.embedding_dim, self.embedding_dim // 3),
                nn.Tanh(),
                nn.Linear(self.embedding_dim // 3, num_heads),  # weight which feature is most important
            )

        # in order to map the output dimennsion of multiple attention heads
        # back to the embedding_dim, we use a linear mapper
        self.mapper = nn.Linear(self.embedding_dim * num_heads, self.embedding_dim)

    def _attenuate(self, x):
        """
        Attentuate one bag of instances (x)
        """
        if self.gated:
            return self.attention(self.attention_v(x) * self.attention_u(x))
        else:
            return self.attention(x)

    def forward(self, x, supercase_indices):
        outs = []
        weights = []
        # since we don't know about the individual bag sizes
        # we iterate over the unique indices and perform the calculations individually
        # This should be optimized for runtime in the future.
        for idxs in supercase_indices.unique():
            case = x[supercase_indices == idxs]
            att = self._attenuate(case)  # Bags x Num Instances x ATTENTION_BRANCHES
            att = torch.transpose(att, 1, 0)  # Bags x ATTENTION_BRANCHES x Num Instances
            att = F.softmax(att, dim=1).clamp(1e-5)  # softmax over instances of one Bag

            outs.append(self.mapper(torch.matmul(att, case).view(1, -1)))
            weights.append(att.view(-1, self.num_heads))

        return torch.cat(outs).reshape(-1, self.embedding_dim), torch.cat(weights).reshape(-1, self.num_heads)


class WeightedAvgPoolReducer(nn.Module):
    """
    Simple weighted average to reduce weight and pool all instances within the Bag
    """

    def __init__(self, embedding_dim) -> None:
        super(WeightedAvgPoolReducer, self).__init__()
        self.embedding_dim = embedding_dim
        self.weightgiver = nn.Linear(self.embedding_dim, 1)

    def average_group_pool(self, subcases: torch.Tensor, supercase_indices: torch.Tensor):
        counts = torch.bincount(supercase_indices)
        supercase_repr = torch.zeros(counts.shape[0], self.embedding_dim, device=subcases.device)
        supercase_repr.index_add_(0, supercase_indices, subcases)
        return supercase_repr / counts.float().unsqueeze(1)

    def forward(self, x, supercase_indices):
        weights = F.sigmoid(self.weightgiver(x))
        # Apply weights to subcases
        x = x * weights
        return self.average_group_pool(x, supercase_indices), weights


class Grouper(SharedBlock):
    class Config(SharedBlock.Config):
        version: Literal["weighted", "attention", "gated-attention"] = "weighted"
        attention_heads: int = 1
        module_name: str = "grouper"

    def __init__(self, args: Config, embedding_dim: int):
        super().__init__(args)
        self.args: Grouper.Config = args
        self.embedding_dim = embedding_dim
        # different versions also need to be treated differently
        if self.args.version == "weighted":
            self.reducer = WeightedAvgPoolReducer(self.embedding_dim)
        elif self.args.version == "attention":
            self.reducer = AttentionPoolingReducer(self.embedding_dim, gated=False, num_heads=self.args.attention_heads)
        elif self.args.version == "gated-attention":
            self.reducer = AttentionPoolingReducer(self.embedding_dim, gated=True, num_heads=self.args.attention_heads)
        else:
            raise NotImplementedError
            "The selected grouper is not implemented"
        self.make_mtl_compatible()

    @staticmethod
    def extract_ids_from_batch(ids: list[str]):
        """
        Takes a list of ids like ["a", "a", "b"] and return a torch.Tensor([0, 0, 1])
        """
        unique_ids = list(set(ids))
        id_to_index = {id_: i for i, id_ in enumerate(unique_ids)}
        return torch.tensor([id_to_index[id_] for id_ in ids])

    def unify_bagsizes(
        self, bag: torch.Tensor, supercase_indices: torch.Tensor, supercase_labels: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        function to unify the sizes of a batch of bags to keep attention based training more efficient
        and keep the flixibility of different bag sizes. It is assumed, that all images have the same CxHxW dimensions
        MIGHT RESULT IN MULTIPLE GRADIENTS FOR ONE IMAGE!
        """
        if self.args.version != "weighted":
            _, c, h, w = bag.shape
            # get supercase infos of how many different bags there are and select the biggest one
            uniques, counts = supercase_indices.unique(return_counts=True)
            if torch.all(counts == counts[0]):
                return bag, supercase_indices, supercase_labels
            else:
                new_bag_size = counts.max()
                # get all instances of each bag
                l = []
                new_indices = []
                new_labels = []
                for u in uniques:
                    idx = torch.argwhere(supercase_indices == u)
                    label = supercase_labels[idx].unique()
                    b_tmp = bag[idx].reshape(-1, c, h, w)
                    # and repeat them new_bag_size_time (complete overkill and need to be optimized)
                    l.append(b_tmp.repeat(new_bag_size, 1, 1, 1)[:new_bag_size])
                    new_indices.append(torch.Tensor([u for _ in range(new_bag_size)]))
                    new_labels.append(torch.Tensor([label for _ in range(new_bag_size)]))
                unified_bag = torch.stack(l).reshape(new_bag_size * len(counts), c, h, w)  # Batch x N x hidden_dim
                new_indices = torch.stack(new_indices).flatten()
                new_labels = torch.stack(new_labels).flatten()
                return (
                    unified_bag,
                    new_indices.long(),
                    new_labels.type(supercase_labels.dtype).to(supercase_labels.device),
                )
        else:
            return bag, supercase_indices, supercase_labels

    def group_targets(self, subcase_labels: torch.Tensor, supercase_indices: torch.Tensor):
        """
        For example in classification, all subcases have the same target.
        In consequence, each subcase label is the correct supercase label.
        """
        # For each supercase we need to identify a subcase where we can extract the label from
        subcase_representative_indices = torch.LongTensor(
            [torch.where(supercase_indices == supercase_index)[0][0] for supercase_index in supercase_indices.unique()]
        )
        return subcase_labels[subcase_representative_indices]

    def forward(self, subcases: torch.Tensor, supercase_indices: torch.Tensor):
        """
        subcases (float): (batch_size, embedding_dim)
        supercase_indices (long): (batch_size,)
        """

        return self.reducer(subcases, supercase_indices)



from __future__ import annotations
from typing import Any, Optional, Callable, Dict, Tuple, TypeVar
from copy import deepcopy

import torch.nn as nn
from torchvision import transforms
from torch.utils.data import Dataset

from .MTLDataset import MTLDataset, SrcCaseType, mtl_batch_collate
from mmm.torch_ext import CachingSubCaseDS
from mmm.data_loading.TrainValCohort import TrainValCohort

CaseDict = TypeVar("CaseDict", bound=Dict)


class ImageGenerationDataset(MTLDataset):
    def __init__(
        self,
        mtl_ds: Dataset[SrcCaseType],
        src_transform: Optional[Callable[[SrcCaseType], Dict[str, Any]]] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mtl_ds, ["image", "target"], ["meta"], src_transform=src_transform, *args, **kwargs)

    def verify_case_by_index(self, index: int) -> Dict[str, Any]:
        case = super().verify_case_by_index(index)
        self.assert_image_data_assumptions(case["image"])
        self.assert_image_data_assumptions(case["target"])
        return case

    def get_input_output_tuple(self, batch: Dict[str, Any]) -> Tuple[Any, ...]:
        return batch["image"], batch["target"]

    def st_case_viewer(self, case: Dict[str, Any], X) -> None:
        import streamlit as st
        from mmm.logging.st_ext import blend_with_mask

        im = case["image"]
        blend_with_mask(im, None, caption_suffix=f"Shape: {im.shape}")
        targetim = case["target"]
        blend_with_mask(targetim, None, caption_suffix=f"Shape: {targetim.shape}")
        st.write(case)

    def _compute_batchsize_from_batch(self, batch: Dict[str, Any]) -> int:
        return batch["image"].shape[0]

    def _visualize_batch_case(self, batch: Dict[str, Any], i: int) -> None:
        import streamlit as st
        from mmm.logging.st_ext import blend_with_mask

        blend_with_mask(
            batch["image"][i],
            None,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}: {batch['image'][i].shape}",
            st_key=f"b{i}",
        )
        blend_with_mask(
            batch["target"][i],
            None,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}: {batch['target'][i].shape}",
            st_key=f"b{i}",
        )


def transform_cohort_to_generative(cohort: TrainValCohort, transform: Callable):
    """
    Transforms any given TrainValCohort into a TrainValCohort with an ImageGenerationDataset.
    transform: Callable should be a  mmm/augmentations.py --> GenerativeTransform,
    which generates a targegt image from the image in the dataset. Will return a dictionary with image and target keys.
    """
    if cohort.datasets[0].src_transform is None:
        cohort.datasets[0].src_transform = nn.Identity()

    new_src_transform = transforms.Compose([cohort.datasets[0].src_transform, transform])

    # Not beautiful but needed.
    if isinstance(cohort.datasets[0].src_ds, CachingSubCaseDS):
        train_ds = cohort.datasets[0].src_ds
        val_ds = cohort.datasets[1].src_ds
    else:
        train_ds = deepcopy(cohort.datasets[0].src_ds)
        val_ds = deepcopy(cohort.datasets[1].src_ds)

    return TrainValCohort(
        args=cohort.args,
        train_ds=ImageGenerationDataset(
            mtl_ds=train_ds,
            src_transform=new_src_transform,
            batch_transform=lambda x: [x]
            if len(x["image"].shape) < 4
            else [
                {"image": subX, "target": subY, "meta": subM}
                for subX, subY, subM in zip(x["image"], x["target"], x["meta"])
            ],
            collate_fn=mtl_batch_collate,
        ),
        val_ds=ImageGenerationDataset(
            mtl_ds=val_ds,
            src_transform=new_src_transform,
            batch_transform=lambda x: [x]
            if len(x["image"].shape) < 4
            else [
                {"image": subX, "target": subY, "meta": subM}
                for subX, subY, subM in zip(x["image"], x["target"], x["meta"])
            ],
            collate_fn=mtl_batch_collate,
        ),
    )



import logging
import cv2
import numpy as np
from typing import Dict, Any, List, TypeVar, Tuple, Optional, Union, Callable
from PIL.Image import Image
import cv2 as cv
import torch
import torch.nn as nn
from torchvision import transforms
from monai.data.meta_tensor import MetaTensor
import monai.transforms as monai_transforms
import albumentations as A
import torchvision.transforms.functional as F
from mmm.BaseModel import BaseModel
from mmm.transforms import RandomApply
from mmm.logging.type_ext import TransformsSeqType
from mmm.settings import mtl_settings

# Typevariable for either a PIL image or a Tensor, indicates that the augmentation does not change the type
ImageType = TypeVar("ImageType", Image, torch.Tensor)

# Taken directly from albumentations.composition (but redefined since it is not exported there)


def get_histo_augs(
    img_fill_value=(255, 255, 255), mask_fill_value=mtl_settings.ignore_class_value
) -> TransformsSeqType:
    return [
        A.OneOf(  # Color variation (OneOf always executes exactly one if it is chosen)
            [
                # Included for positive transfer to grayscale images via multi-task learning
                A.ToGray(p=0.2),
                A.ChannelShuffle(p=0.2),
                # Should approximate typical color variation in stains
                A.HueSaturationValue(
                    hue_shift_limit=30, sat_shift_limit=20, val_shift_limit=20, always_apply=False, p=0.7
                ),
            ],
            p=0.5,
        ),
        A.OneOf(
            [
                A.CLAHE(clip_limit=2),
                A.Sharpen(),
                A.Emboss(),
                A.RandomBrightnessContrast(),
            ],
            p=0.5,
        ),
        A.OneOf(
            [
                A.RandomGamma(p=1),
                A.GaussNoise(),
            ],
            p=0.3,
        ),
        A.OneOf(
            [
                A.MotionBlur(p=0.2),
                A.MedianBlur(blur_limit=3, p=0.1),
                A.Blur(blur_limit=3, p=0.1),
            ],
            p=0.3,
        ),
        A.HorizontalFlip(p=0.5),
        A.VerticalFlip(p=0.5),
        A.RandomRotate90(p=0.5),
        A.ShiftScaleRotate(
            rotate_limit=10,
            border_mode=cv2.BORDER_CONSTANT,
            value=img_fill_value,
            mask_value=mask_fill_value,
        ),
    ]


def get_artefact_histo_augs():
    return [
        RandomBlobAugmentation(),
        RandomLineAugmentation(),
        RandomBlobAugmentation(),
    ]


def get_weak_default_augs(img_fill_value=0, mask_fill_value=mtl_settings.ignore_class_value) -> TransformsSeqType:
    """
    Designed to be compatible with as many as possible intensity-based images.
    Does not flip.
    """
    return [
        A.OneOf(
            [
                A.RandomGamma(p=1),
                A.GaussNoise(),
            ],
            p=0.1,
        ),
        A.ShiftScaleRotate(
            shift_limit=0.0625,
            scale_limit=0.1,
            rotate_limit=10,
            p=0.8,
            border_mode=cv2.BORDER_CONSTANT,
            value=img_fill_value,
            mask_value=mask_fill_value,
        ),
        A.OneOf(
            [
                A.OpticalDistortion(p=0.3),
                A.GridDistortion(p=0.1),
                # This has a bug, it swaps all non-zero labels in the mask to one :O
                # A.PiecewiseAffine(p=0.3),
            ],
            p=0.1,
        ),
        A.OneOf(
            [
                A.CLAHE(clip_limit=2),
                A.Sharpen(),
                A.Emboss(),
                A.RandomBrightnessContrast(),
            ],
            p=0.2,
        ),
        A.HueSaturationValue(p=0.2),
    ]


def get_realworld_augs() -> TransformsSeqType:
    return [
        A.RandomRotate90(),
        A.Flip(),
        A.Transpose(),
        A.ToGray(p=0.2),
        A.OneOf(
            [
                # A.IAAAdditiveGaussianNoise(),
                A.RandomGamma(p=1),
                A.GaussNoise(),
            ],
            p=0.2,
        ),
        A.OneOf(
            [
                A.MotionBlur(p=0.2),
                A.MedianBlur(blur_limit=3, p=0.1),
                A.Blur(blur_limit=3, p=0.1),
            ],
            p=0.2,
        ),
        A.ShiftScaleRotate(shift_limit=0.0625, scale_limit=0.2, rotate_limit=45, p=0.2),
        A.OneOf(
            [
                A.OpticalDistortion(p=0.3),
                A.GridDistortion(p=0.1),
                # This has a bug, it swaps all non-zero labels in the mask to one :O
                # A.PiecewiseAffine(p=0.3),
            ],
            p=0.2,
        ),
        A.OneOf(
            [
                A.CLAHE(clip_limit=2),
                A.Sharpen(),
                A.Emboss(),
                A.RandomBrightnessContrast(),
            ],
            p=0.3,
        ),
        A.HueSaturationValue(p=0.3),
    ]


def get_xray_augs(img_fill_value=0, mask_fill_value=mtl_settings.ignore_class_value) -> TransformsSeqType:
    return [
        A.InvertImg(p=0.3),
        A.OneOf(
            [
                A.CLAHE(clip_limit=2),
                A.Sharpen(),
                A.Emboss(),
                A.RandomBrightnessContrast(),
            ],
            p=0.3,
        ),
        A.OneOf(
            [
                # A.IAAAdditiveGaussianNoise(),
                A.RandomGamma(p=1),
                A.GaussNoise(),
            ],
            p=0.2,
        ),
        A.OneOf(
            [
                A.MotionBlur(p=0.2),
                A.MedianBlur(blur_limit=3, p=0.1),
                A.Blur(blur_limit=3, p=0.1),
            ],
            p=0.2,
        ),
        # A.OneOf([
        #     A.OpticalDistortion(p=0.3),
        #     A.GridDistortion(p=.1),
        #     A.PiecewiseAffine(p=0.3),
        # ], p=0.2),
        A.ShiftScaleRotate(
            p=1.0,
            rotate_limit=10,
            border_mode=cv2.BORDER_CONSTANT,
            value=img_fill_value,
            mask_value=mask_fill_value,
        ),
    ]


def get_mri2d_augs() -> TransformsSeqType:
    return [
        A.RandomRotate90(),
        A.Flip(),
        A.Transpose(),
        A.OneOf(
            [
                A.RandomGamma(p=1),
                A.GaussNoise(),
            ],
            p=0.2,
        ),
        A.OneOf(
            [
                A.MotionBlur(p=0.2),
                A.MedianBlur(blur_limit=3, p=0.1),
                A.Blur(blur_limit=3, p=0.1),
            ],
            p=0.2,
        ),
        A.ShiftScaleRotate(
            shift_limit=0.0625,
            scale_limit=0.2,
            rotate_limit=45,
            p=0.2,
            border_mode=cv2.BORDER_CONSTANT,
        ),
        A.OneOf(
            [
                A.OpticalDistortion(p=0.3),
                A.GridDistortion(p=0.1),
                # This has a bug, it swaps all non-zero labels in the mask to one :O
                # A.PiecewiseAffine(p=0.3),
            ],
            p=0.2,
        ),
        A.OneOf(
            [
                A.CLAHE(clip_limit=2),
                A.Sharpen(),
                A.Emboss(),
                A.RandomBrightnessContrast(),
            ],
            p=0.3,
        ),
        A.HueSaturationValue(p=0.3),
    ]


def get_contrastive_2D_augs():
    """
    Contrastive augmentations based on https://arxiv.org/pdf/2105.04906.pdf
    """
    return [
        transforms.RandomResizedCrop(224, scale=(0.08, 0.1)),
        transforms.RandomHorizontalFlip(p=0.5),
        transforms.RandomApply([transforms.ColorJitter(0.4, 0.4, 0.2, 0.1)], p=0.8),
        transforms.RandomGrayscale(p=0.2),
        # transforms.RandomSolarize(p=0.1, threshold=), # left our bcs I don't know about the threshold yes
        transforms.RandomApply([transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))]),
    ]


def volume_scale_to_0_1(img: torch.Tensor):
    """
    Scales the input array to the range [0, 1].
    """
    # Make sure extreme outliers do not influence the scaling
    return monai_transforms.ScaleIntensityRangePercentiles(lower=2.5, upper=97.5, b_min=0.0, b_max=1.0, clip=True)(img)


class MRI3DProcessor:
    class Config(BaseModel):
        normalize: bool = True
        rotate_3d: bool = False
        augment_intensity: bool = False

    @staticmethod
    def get_3d_rotations(keys):
        """
        If the shortest axis is longer than half the longest axis 3d rotations are a good idea.
        """
        return [
            monai_transforms.RandRotate90D(keys=keys, prob=0.2, spatial_axes=(0, 1)),
            monai_transforms.RandRotate90D(keys=keys, prob=0.2, spatial_axes=(0, 2)),
            monai_transforms.RandRotate90D(keys=keys, prob=0.2, spatial_axes=(1, 2)),
        ]

    @staticmethod
    def base_volume_augs(keys):
        return [
            monai_transforms.RandAdjustContrastD(keys=["image"], prob=0.1),
            RandomApply(
                monai_transforms.RandScaleCropD(keys=keys, roi_scale=[0.8, 0.8, 0.8], random_size=True),
                p=0.3,
            ),
        ]

    def __init__(self, args, augs_constructor: Optional[Callable], with_segmask: bool) -> None:
        self.args = args
        if augs_constructor:
            self.augs = monai_transforms.Compose(
                augs_constructor(["image", "label"] if with_segmask else ["image"])
                # self.get_recommended_mri_augs(keys=["image", "label"] if with_segmask else ["image"])
            )
        else:
            self.augs = None

    def __call__(self, d: Dict[str, Any]) -> Dict[str, Any]:
        assert "image" in d and len(d["image"].shape) >= 4, "C, spatials... dimensions expected"
        if "label" in d:
            assert len(d["image"].shape) == len(d["label"].shape) + 1

            # Monai expects a channel dimension which is always empty for multiclass problems
            d["label"] = torch.unsqueeze(d["label"], 0)

        if self.args.normalize:
            d["image"] = volume_scale_to_0_1(d["image"])
        if self.augs is not None:
            d = self.augs(d)  # type: ignore

        # Get rid of the MONAI meta tensor
        if isinstance(d["image"], MetaTensor):
            d["image"] = d["image"].as_tensor()

        if "label" in d:
            if isinstance(d["label"], MetaTensor):
                d["label"] = d["label"].as_tensor()
            d["label"] = torch.squeeze(d["label"])
        return d


def _case_to_aformat(
    d: Dict[str, Any]
) -> Tuple[np.ndarray, Optional[np.ndarray], Optional[List[float]], Optional[List[int]]]:
    aimg = (d["image"] * 255.0).permute(1, 2, 0).numpy().astype(np.uint8)
    amask = d["label"].numpy() if "label" in d else None
    aboxes = d["boxes"].type(torch.LongTensor).tolist() if "boxes" in d else None
    alabels = d["labels"].tolist() if "labels" in d else None
    return aimg, amask, aboxes, alabels


def _atransform_into_case(transformed: Dict[str, Any]) -> Dict[str, Any]:
    res = {"image": F.to_tensor(transformed["image"])}  # torch.from_numpy(transformed['image']).permute(2, 0, 1),
    if "bboxes" in transformed:
        res["boxes"] = torch.Tensor(transformed["bboxes"])
        res["labels"] = torch.Tensor(transformed["class_labels"]).long()
    if "mask" in transformed:
        res["label"] = torch.from_numpy(transformed["mask"]).long()
    return res


class Alb:
    def __init__(self, transforms: TransformsSeqType, support_boxes=False) -> None:
        box_kwargs = (
            {"bbox_params": A.BboxParams(format="pascal_voc", label_fields=["class_labels"])} if support_boxes else {}
        )
        self.transform = A.Compose(
            transforms,
            **box_kwargs,
        )

    def __call__(self, d: Dict[str, Any]) -> Dict[str, Any]:
        aimg, amask, aboxes, alabels = _case_to_aformat(d)

        transform_kwargs: Dict[str, Any] = dict(image=aimg)
        for kwarg_name, kwarg_value in [
            ("mask", amask),
            ("bboxes", aboxes),
            ("class_labels", alabels),
        ]:
            if kwarg_value is not None:
                transform_kwargs[kwarg_name] = kwarg_value
        transformed = self.transform(**transform_kwargs)

        transformed_case = _atransform_into_case(transformed)
        d.update(transformed_case)
        return d


class AlbWithBoxes(Alb):
    def __init__(self, transforms: TransformsSeqType) -> None:
        super().__init__(transforms, support_boxes=True)


class PILHistoPatchAug:
    def __init__(self):
        self.f = transforms.Compose(
            [
                transforms.RandomHorizontalFlip(p=0.5),
                transforms.RandomVerticalFlip(p=0.5),
                transforms.ColorJitter(brightness=0.2, contrast=0.75, saturation=0.25, hue=0.04),
                transforms.RandomAffine(
                    degrees=10,
                    translate=(0.01, 0.01),
                    scale=(0.9, 1.3),
                    shear=(-0.1, 0.1),
                    fill=255,
                ),
            ]
        )

    def __call__(self, im: ImageType) -> ImageType:
        return self.f(im)


class SimCLRPatchAug:
    def __init__(self):
        self.cl_transform = transforms.Compose(
            [
                transforms.RandomHorizontalFlip(p=0.5),
                transforms.RandomVerticalFlip(p=0.5),
                transforms.RandomApply(
                    [transforms.GaussianBlur(kernel_size=(7, 7), sigma=(0.1, 0.2))],
                    p=0.5,
                ),
                transforms.RandomApply(
                    [transforms.ColorJitter(contrast=0.8, brightness=0.8, saturation=0.8, hue=0.2)],
                    p=0.8,
                ),
                transforms.RandomAffine(
                    degrees=90,
                    translate=(0.01, 0.01),
                    scale=(0.9, 1.3),
                    shear=(-0.1, 0.1),
                    fill=255,
                ),
                transforms.RandomGrayscale(p=0.2),
                transforms.ToTensor(),
            ]
        )

    def __call__(self, img: ImageType) -> ImageType:
        return self.cl_transform(img)


class RandomLineAugmentation:
    def __init__(self, p=0.7) -> None:
        self.p = p

    def __call__(self, image: torch.Tensor):
        """
        Expect a normlized RGB Torch Tensor
        """
        local_p = torch.rand(1).item()
        if local_p > self.p:
            return image
        else:
            data_c, data_x, data_y = image.squeeze().shape
            # generate random points (without replacement)
            rnd_x = torch.randperm(data_x)[:2]
            rnd_y = torch.randperm(data_y)[:2]

            # transfrom into cv2 readable landmarks
            pt1 = (rnd_x[0].item(), rnd_y[0].item())
            pt2 = (rnd_x[1].item(), rnd_y[1].item())

            # create artifact mask
            artifacts = torch.ones((data_x, data_y, 1), dtype=torch.uint8).numpy() * 255
            line_mask = cv.line(artifacts, pt1, pt2, color=(0, 0, 0), thickness=data_x // 32) / 255

            rnd_blurr_divider = torch.randint(15, 20, size=(1,)).item()
            # (data_x//rnd_blurr_divider, data_y//rnd_blurr_divider))
            blurred_mask = cv.blur(line_mask, (data_x // rnd_blurr_divider, data_y // rnd_blurr_divider))

            # multiply input img with mask to create artifact
            return image * blurred_mask  # , blurred_mask


class RandomBlobAugmentation:
    def __init__(self, p=0.5) -> None:
        self.p = p

    def __call__(self, image: torch.Tensor) -> ImageType:
        local_p = torch.rand(1).item()
        if local_p > self.p:
            return image
        else:
            data_c, data_x, data_y = image.squeeze().shape
            # generate random points
            min_radius = data_x // 30
            rnd_x = torch.randint(min_radius, data_x - min_radius, size=(1,))
            rnd_y = torch.randint(min_radius, data_y - min_radius, size=(1,))
            center = (rnd_x.item(), rnd_y.item())
            rnd_size_divider = torch.randint(27, 50, size=(1,)).item()

            artifacts = torch.ones((data_x, data_y, 1), dtype=torch.uint8).numpy() * 255
            circle_mask = (
                cv.circle(
                    artifacts,
                    center,
                    radius=data_x // rnd_size_divider,
                    color=(0, 0, 0),
                    thickness=-1,
                )
                / 255
            )
            rnd_blurr_divider = torch.randint(15, 20, size=(1,)).item()
            blurred_mask = cv.blur(circle_mask, (data_x // rnd_blurr_divider, data_y // rnd_blurr_divider))

            return image * blurred_mask  # , blurred_mask


class GenerativeTransform:
    """
    Designed to be used as target transform in the GenerativeDataset.
    Receives two callable transforms, which modify the image input image.
    """

    def __init__(self, image_transform: Callable | None, target_transform: Callable | None) -> None:
        self.it = image_transform
        self.tt = target_transform

    def __call__(self, case: dict[str, torch.Tensor]) -> Any:
        res = {
            "image": self.it(case["image"]) if self.it else case["image"],
            "target": self.tt(case["image"]) if self.tt else case["image"],
            "meta": case["meta"] if "meta" in list(case.keys()) else {},
        }
        return res


def get_color_prediction_transform():
    """
    Reconstructing a normalized version of the input image
    Standard ImageNet-based normalization used for RGB images
    if working with 1 channel images, other normalzation should be used
    """
    return GenerativeTransform(
        image_transform=None,
        target_transform=transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
    )


def get_surrounding_prediction_transform(center_crop: tuple[int]):
    """
    Reconstruction of the image itself and the surroundings.
    To be used with the SEMSEG decoder or attention head.
    """
    return GenerativeTransform(
        image_transform=transforms.RandomResizedCrop(size=center_crop),
        target_transform=transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
    )



"""
Trainable simulated vision tasks with a low loading time and memory requirements
for automatic testing.
"""

from __future__ import annotations
from enum import Enum
from itertools import product

import random
from typing import Callable, Dict, Tuple, Any, List, Type, Union, Literal
import numpy as np
from PIL.Image import Image, new as pil_new
from PIL.ImageDraw import ImageDraw, Draw
from pydantic import Field

import torchvision.transforms as torch_transforms

import torch
from torch.utils.data import Dataset, IterableDataset, get_worker_info
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.BaseModel import BaseModel
from mmm.bucketizing import Bucket, BucketConfig
from mmm.data_loading.utils import TransformedSubset
from mmm.transforms import ApplyToKey, KeepOnlyKeysInDict
from mmm.data_loading.SemSegDataset import SemSegDataset
from mmm.data_loading.DetectionDataset import DetectionDataset
from mmm.data_loading.ClassificationDataset import ClassificationDataset
from mmm.data_loading.ImageCaptionDataset import ImageCaptionDataset
from mmm.data_loading.MultilabelClassificationDataset import (
    MultilabelClassificationDataset,
)
from mmm.data_loading.MultiLabelSemSegDataset import MultiLabelSemSegDataset


def get_random_bounds(min_max_size=(0.1, 0.3), canvas_size=(224, 224)) -> Tuple[int, int, int, int]:
    width = random.randint(int(canvas_size[0] * min_max_size[0]), int(canvas_size[0] * min_max_size[1]))
    height = random.randint(int(canvas_size[1] * min_max_size[0]), int(canvas_size[1] * min_max_size[1]))
    x = random.randint(0, canvas_size[0] - width)
    y = random.randint(0, canvas_size[1] - height)
    return x, y, x + width, y + height


def track_difference(img: Image, drawing_function: Callable[[ImageDraw], Any]):
    """
    Given an existing image and the current mask, draws a new object and adds the difference image to the mask.
    """
    original_im = np.array(img)
    draw = Draw(img)
    drawing_function(draw)

    minus_im = np.array(img) != (original_im)

    # If a pixel is changed in any colorchannel, the mask gets that value
    changed_pixels = np.logical_or(*[minus_im[:, :, channel_dim] for channel_dim in range(minus_im.shape[2])])

    return img, changed_pixels


class CanvasConfig(BaseModel):
    canvas_size: Tuple[int, int] = (512, 512)
    bg_color: Tuple[int, int, int] = (128, 128, 128)
    obj_size_range: Tuple[float, float] = (0.1, 0.3)


class ShapeColor(str, Enum):
    Charcoal = "28536B"
    RosyBrown = "C2948A"
    Khaki = "BBB193"
    TurkeyRed = "A30000"
    Asparagus = "659157"
    Moonstone = "69A2B0"

    @staticmethod
    def hex_to_rgb(hex: str) -> Tuple[int, int, int]:
        if hex.startswith("#"):
            hex = hex[1:]
        return tuple(int(hex[i : i + 2], 16) for i in (0, 2, 4))

    @staticmethod
    def toRGB(c: ShapeColor) -> Tuple[int, int, int]:
        return ShapeColor.hex_to_rgb(c.value)


class BaseObject(BaseModel):
    description: str
    in_area: Tuple[int, int, int, int]

    @staticmethod
    def sample_object_for_canvas(settings: CanvasConfig) -> BaseObject:
        raise NotImplementedError

    def draw_shape(self, canvas: ImageDraw) -> ImageDraw:
        raise NotImplementedError


class Ellipse(BaseObject):
    obj_type: Literal["ellipse"] = "ellipse"
    fillcolor: ShapeColor

    @staticmethod
    def sample_object_for_canvas(settings: CanvasConfig) -> BaseObject:
        fillcolor = random.choice(list(ShapeColor))
        outer_area = get_random_bounds(min_max_size=settings.obj_size_range, canvas_size=settings.canvas_size)
        return Ellipse(
            description=f"A {fillcolor.name.lower()} ellipse",
            in_area=outer_area,
            fillcolor=fillcolor,
        )

    def draw_shape(self, canvas: ImageDraw) -> ImageDraw:
        canvas.ellipse(self.in_area, fill=ShapeColor.toRGB(self.fillcolor))
        return canvas


class Rectangle(BaseObject):
    obj_type: Literal["rectangle"] = "rectangle"
    fillcolor: ShapeColor

    @staticmethod
    def sample_object_for_canvas(settings: CanvasConfig) -> BaseObject:
        fillcolor = random.choice(list(ShapeColor))
        outer_area = get_random_bounds(min_max_size=settings.obj_size_range, canvas_size=settings.canvas_size)
        return Rectangle(
            description=f"A {fillcolor.name.lower()} rectangle",
            in_area=outer_area,
            fillcolor=fillcolor,
        )

    def draw_shape(self, canvas: ImageDraw) -> ImageDraw:
        canvas.rectangle(self.in_area, fill=ShapeColor.toRGB(self.fillcolor))
        return canvas


ObjectTypes = BaseObject.__subclasses__()


class ShapeDataset(Dataset[Dict[str, Any]]):
    """
    Dataset which uses PIL to draw simple shapes.

    `get_class_names()`'s index corresponds to the value in the segmentation masks.
    """

    class Config(BaseModel):
        N: int = 10
        canvas: CanvasConfig = CanvasConfig()
        num_objects_per_scene_range: Tuple[int, int] = (3, 10)
        # canvas_buckets: BucketConfig
        object_buckets: BucketConfig = BucketConfig(
            exactly_one_match=True,
            buckets=[
                Bucket(expression=".*ell.*", description="ellipse"),
                Bucket(expression=".*rect.*", description="rectangle"),
            ],
        )

    def __init__(self, c: Config) -> None:
        self.config = c

        # Each case stores several objects
        self.case_list = [self.generate_objects(self.config.num_objects_per_scene_range) for _ in range(self.config.N)]

    def get_class_names(self) -> List[str]:
        # return ["Background"] + [" ".join([color.value, shape.value])
        #                          for color, shape in product(ShapeColor, ShapeType)]
        return self.config.object_buckets.get_class_names()

    def generate_objects(self, num_objects_range: Tuple[int, int]) -> List[BaseObject]:
        num_objects = random.randint(num_objects_range[0], num_objects_range[1])
        object_types: List[Type[BaseObject]] = [random.choice(ObjectTypes) for _ in range(num_objects)]

        # coords = [get_random_bounds(
        #     min_max_size=self.config.canvas.obj_size_range, canvas_size=self.config.canvas.canvas_size
        # ) for _ in range(num_objects)]
        objects = [o.sample_object_for_canvas(self.config.canvas) for o in object_types]

        return objects

    @staticmethod
    def cohort_transform():
        return torch_transforms.Compose(
            [
                ApplyToKey(torch_transforms.ToTensor(), "image"),
                ApplyToKey(torch.LongTensor, "label"),
                ApplyToKey(torch.LongTensor, "masks"),
                ApplyToKey(torch.FloatTensor, "boxes"),
                ApplyToKey(torch.LongTensor, "labels"),
            ]
        )

    def __iter__(self):
        for objects in self.case_list:
            yield self._generate_case_for_objects(objects)

    def __len__(self) -> int:
        return self.config.N

    def _generate_case_for_objects(self, scene_elements: List[BaseObject]) -> Dict[str, Any]:
        im: Image = pil_new("RGB", self.config.canvas.canvas_size[::-1], self.config.canvas.bg_color)
        mask = np.zeros(shape=self.config.canvas.canvas_size, dtype=np.int64)
        masks = np.zeros(
            shape=[len(self.config.object_buckets.get_class_names()) + 1] + list(self.config.canvas.canvas_size),
            dtype=np.int64,
        )
        masks[0, ...] = 1
        boxes = []
        box_labels = []
        class_names = []
        for obj in scene_elements:
            # element_label = self._mask_lookup[(shape, color)]
            class_name, element_label = self.config.object_buckets.get_bucket_name(obj.description)

            im, changed_pixels = track_difference(im, obj.draw_shape)
            mask[changed_pixels] = element_label + 1
            masks[element_label + 1, changed_pixels] = 1  # Set object at changed_pixels to 1
            masks[0, changed_pixels] = 0  # Set background at changed_pixels to 0
            boxes.append(obj.in_area)
            box_labels.append(element_label)
            class_names.append(class_name)

        class_labels = [float(class_index in box_labels) for class_index, _ in enumerate(self.get_class_names())]
        return {
            "image": im,
            "label": mask,
            "masks": masks,
            "boxes": boxes,
            "labels": box_labels,
            "class_labels": torch.Tensor(class_labels).float(),
            "obj_class_names": class_names,
            "meta": {
                "elements": scene_elements,
            },
        }

    def __getitem__(self, i: int) -> Dict[str, Any]:
        scene_elements: List[BaseObject] = self.case_list[i]
        return self._generate_case_for_objects(scene_elements)


def build_caption(case: dict[str, Any]):
    """Assumes that case["meta"]["elements"] exists"""
    if case["meta"]["elements"]:
        case["captions"] = [
            " and ".join([e.description for e in case["meta"]["elements"]]),
            "A scene with " + ", ".join([e.description for e in case["meta"]["elements"]]),
        ]
    else:
        case["captions"] = ["No objects in scene", "Empty scene"]
    return case


class IterShapeDataset(IterableDataset[Dict[str, Any]], ShapeDataset):
    def __iter__(self):
        if (worker_info := get_worker_info()) is not None:
            if worker_info.num_workers > 1:
                raise Exception(f"Iterable shape dataset cannot be used with more than 1 worker.")
        for objects in self.case_list:
            yield self._generate_case_for_objects(objects)


if __name__ == "__main__":
    from mmm.interactive import configs as cfs
    import streamlit as st
    from mmm.logging.st_ext import multi_cohort_explorer, stw

    @st.cache_resource()
    def ds(iterable: bool = False):
        ds_type = IterShapeDataset if iterable else ShapeDataset
        return ds_type(
            ShapeDataset.Config(
                N=1000,
                num_objects_per_scene_range=(1, 3),
                canvas=CanvasConfig(canvas_size=(224, 224)),
            )
        ), ShapeDataset(
            ShapeDataset.Config(
                N=500,
                num_objects_per_scene_range=(1, 1),
                canvas=CanvasConfig(canvas_size=(224, 224)),
            )
        )

    train_ds, val_ds = ds()

    def build_clf_cohort() -> TrainValCohort[ClassificationDataset]:
        classnames = ["No rectangle exists", "Rectangle exists"]

        def transform_for_label(case: Dict[str, Any]) -> Dict[str, Any]:
            case["class"] = int("rectangle" in case["obj_class_names"])
            return case

        base_transform = torch_transforms.Compose([ShapeDataset.cohort_transform(), transform_for_label])
        return TrainValCohort(
            cfs.TrainValCohortConfig(batch_size=8, num_workers=0),
            ClassificationDataset(train_ds, src_transform=base_transform, class_names=classnames),
            ClassificationDataset(val_ds, src_transform=base_transform, class_names=classnames),
        )

    def build_ae_cohort():
        from mmm.interactive import pipes, data
        import albumentations as A
        import torchvision.transforms as transforms

        base_t = [
            ShapeDataset.cohort_transform(),
            pipes.CopyKeysInDict({"image": "targetimage"}),
        ]
        augs = pipes.AlbWithBoxes(
            [
                A.RandomFog(p=0.33),
                A.Rotate(limit=45),
                A.RandomRain(p=0.33),
                A.RandomShadow(p=0.33),
                # A.RandomToneCurve(p=0.33),
                A.RandomSnow(p=0.33),
            ]
        )
        return data.TrainValCohort(
            cfs.TrainValCohortConfig(batch_size=8, num_workers=0),
            data.ImageTranslationDataset(train_ds, src_transform=transforms.Compose(base_t + [augs])),
            data.ImageTranslationDataset(val_ds, src_transform=transforms.Compose(base_t + [augs])),
        )

    def build_multilabelclf_cohort() -> TrainValCohort[MultilabelClassificationDataset]:
        base_transform = torch_transforms.Compose([ShapeDataset.cohort_transform()])
        return TrainValCohort(
            cfs.TrainValCohortConfig(batch_size=8, num_workers=0),
            MultilabelClassificationDataset(
                train_ds,
                src_transform=base_transform,
                class_names=train_ds.get_class_names(),
            ),
            MultilabelClassificationDataset(
                val_ds,
                src_transform=base_transform,
                class_names=val_ds.get_class_names(),
            ),
        )

    def build_semseg_cohort() -> TrainValCohort[SemSegDataset]:
        return TrainValCohort(
            cfs.TrainValCohortConfig(batch_size=8, num_workers=0),
            SemSegDataset(
                train_ds,
                src_transform=ShapeDataset.cohort_transform(),
                class_names=["background"] + train_ds.get_class_names(),
            ),
            SemSegDataset(
                val_ds,
                src_transform=ShapeDataset.cohort_transform(),
                class_names=["background"] + val_ds.get_class_names(),
            ),
        )

    def build_det_cohort() -> TrainValCohort[DetectionDataset]:
        return TrainValCohort(
            cfs.TrainValCohortConfig(batch_size=8, num_workers=0),
            DetectionDataset(
                train_ds,
                src_transform=ShapeDataset.cohort_transform(),
                class_names=train_ds.get_class_names(),
            ),
            DetectionDataset(
                val_ds,
                src_transform=ShapeDataset.cohort_transform(),
                class_names=val_ds.get_class_names(),
            ),
        )

    def build_caption_cohort() -> TrainValCohort[ImageCaptionDataset]:
        return TrainValCohort(
            TrainValCohort.Config(batch_size=8, num_workers=0),
            ImageCaptionDataset(
                train_ds,
                src_transform=torch_transforms.Compose([ShapeDataset.cohort_transform(), build_caption]),
            ),
            ImageCaptionDataset(
                val_ds,
                src_transform=torch_transforms.Compose([ShapeDataset.cohort_transform(), build_caption]),
            ),
        )

    def build_multilabelsemseg_cohort() -> TrainValCohort[MultiLabelSemSegDataset]:
        return TrainValCohort(
            cfs.TrainValCohortConfig(batch_size=8, num_workers=0),
            MultiLabelSemSegDataset(
                train_ds,
                src_transform=ShapeDataset.cohort_transform(),
                class_names=["background"] + train_ds.get_class_names(),
            ),
            MultiLabelSemSegDataset(
                val_ds,
                src_transform=ShapeDataset.cohort_transform(),
                class_names=["background"] + val_ds.get_class_names(),
            ),
        )

    multi_cohort_explorer(
        {
            "mlsemseg": build_multilabelsemseg_cohort,
            "caption": build_caption_cohort,
            "mlclf": build_multilabelclf_cohort,
            "clf": build_clf_cohort,
            "ae": build_ae_cohort,
            "shapesemseg": build_semseg_cohort,
            "shapedetection": build_det_cohort,
        }
    )



import torch.nn as nn

from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock

from .MTLTask import MTLTask


class TaskModule(nn.Module):
    """
    Wraps a task along with its shared MTLModule dependencies
    to create a single neural network for other libraries to understand.

    You should make sure the shared modules are using this task (set_active_task) before executing the forward pass.
    """

    def __init__(self, task: MTLTask, shared_modules: dict[str, SharedBlock]) -> None:
        super().__init__()
        self.task = task
        self.shared_modules = nn.ModuleDict(shared_modules)

    def forward(self, x):
        result = self.task.forward(x, self.shared_modules)
        return result



from pathlib import Path
import os
from typing import Optional
from enum import Enum
import numpy as np
import imageio
import wandb
import re

try:
    import plotly.graph_objects as go
    import plotly.express as px
except ImportError:
    go, px = None, None
import torch
import torch.nn as nn

from typing import List, Dict, Callable

import cv2

from mmm.DataSplit import DataSplit
from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.mtl_modules.shared_blocks.PyramidEncoder import PyramidEncoder


def remove_wandb_special_chars(s: str) -> str:
    for special_char in [".", "/"]:
        s = s.replace(special_char, "")
    return s


def convert_mtlimage_to_cv2(im: torch.Tensor) -> np.ndarray:
    """
    CV2 expects channels last for color and no channels at all for gray.
    """
    if im.shape[0] == 1:
        # Gray
        return im[0].numpy()
    else:
        # Our images are channels first, cv2 works with channels last
        channels_last_im = np.moveaxis(im.numpy(), 0, -1)
        return channels_last_im
        # raise NotImplementedError


def convert_cv2_to_mtlimage(cv2_im: np.ndarray) -> torch.Tensor:
    """
    Only works for 2D.
    """
    if len(cv2_im.shape) == 2:
        return torch.from_numpy(cv2_im).unsqueeze(0)
    else:
        # Color image
        channels_first_im = np.moveaxis(cv2_im, -1, 0)
        return torch.from_numpy(channels_first_im)


def resize_if_necessary(im: torch.Tensor, max_spatial_size: int) -> torch.Tensor:
    assert len(im.shape) == 3, "resizing only works for 2D"

    max_edge_length = max(im.shape[1], im.shape[2])
    if max_edge_length > max_spatial_size:
        factor: float = max_spatial_size / max_edge_length
        resized_np = cv2.resize(
            convert_mtlimage_to_cv2(im),
            (0, 0),
            fx=factor,
            fy=factor,
            interpolation=cv2.INTER_AREA,
        )
        return convert_cv2_to_mtlimage(resized_np)
    else:
        return im


def resize_3d_if_necessary(im: torch.Tensor, max_spatial_size: int) -> torch.Tensor:
    assert len(im.shape) == 4, "resizing only works for 3D"
    max_edge_length = max(im.shape[1], im.shape[2])
    if max_edge_length > max_spatial_size:
        factor: float = max_spatial_size / max_edge_length

        slices = [
            cv2.resize(
                convert_mtlimage_to_cv2(im[..., slice_id]),
                (0, 0),
                fx=factor,
                fy=factor,
                interpolation=cv2.INTER_AREA,
            )
            for slice_id in range(im.shape[-1])
        ]

        return torch.stack([torch.from_numpy(x) for x in slices], dim=-1)
    else:
        return im


def build_wandb_image(im: torch.Tensor, caption: str, resample_to_spatial_size: Optional[int] = None) -> wandb.Image:
    if len(im.shape) == 3:
        # 2D images
        im = resize_if_necessary(im, resample_to_spatial_size) if resample_to_spatial_size is not None else im
        return wandb.Image(im, caption=caption)
    elif (len(im.shape)) == 4:
        # Plot a gif over the last dimension
        tmp_dir = Path(wandb.run.dir) / "volumes"
        tmp_dir.mkdir(exist_ok=True)
        tmp_path = tmp_dir / f"out_{len(os.listdir(tmp_dir))}.gif"
        im = resize_3d_if_necessary(im, resample_to_spatial_size) if resample_to_spatial_size is not None else im
        volume = np.moveaxis(np.squeeze(im.numpy() * 255).astype("uint8"), -1, 0)
        imageio.mimsave(tmp_path, volume)
        return wandb.Image(str(tmp_path), caption=caption)

    raise Exception(f"Image is of unexpected shape: {im.shape}")


def build_wandb_image_for_clf(
    im: torch.Tensor,
    gt: np.ndarray,
    pred: np.ndarray,
    class_names: List[str],
    resample_to_spatial_size: Optional[int] = None,
    caption_suffix="",
):
    description = f"{np.min(im.numpy()):.3f}, {np.max(im.numpy()):.3f}; {im.size()}"
    true_str = f"{class_names[gt]} - {gt}"
    pred_str = f"{class_names[pred]} - {pred}"

    wandb_img = build_wandb_image(
        im=im,
        caption=f"{description}\nTrue: {true_str}\nPred: {pred_str}\n{caption_suffix}",
        resample_to_spatial_size=resample_to_spatial_size,
    )

    return wandb_img, description, true_str, pred_str


class PlotModuleMode(Enum):
    full, aggregated, summary = "full", "aggregated", "summary"


def get_module_state(m: nn.Module, comparison_state=None) -> Dict:
    res = {}

    for parameter_name, parameter_tensor in m.named_parameters():
        # clone might not be necessary, but lets be safe until that assumption is tested
        res[f"{parameter_name}"] = parameter_tensor.detach().cpu().clone().flatten()
        if comparison_state and parameter_name in comparison_state:
            res[f"{parameter_name}_change"] = res[f"{parameter_name}"] - comparison_state[parameter_name]

        if parameter_tensor.grad is not None:
            # clone might not be necessary, but lets be safe until that assumption is tested
            res[f"{parameter_name}_grad"] = parameter_tensor.grad.cpu().clone().flatten()
            if comparison_state and f"{parameter_name}_grad" in comparison_state:
                res[f"{parameter_name}_grad_change"] = res[f"{parameter_name}_grad"] - comparison_state[parameter_name]

    return res


def convert_module_state_to_wandb(d: Dict, mode: PlotModuleMode) -> Dict:
    res = {}
    if mode is PlotModuleMode.full:
        for param_name, param_value in d.items():
            if isinstance(param_value, torch.Tensor):
                res[param_name] = wandb.Histogram(param_value)  # type: ignore
            else:
                res[param_name] = param_value
    elif mode is PlotModuleMode.aggregated:
        buffer: Dict[str, List] = {
            "params": [],
            "grads": [],
            "param_changes": [],
            "grad_changes": [],
        }

        aggregator = torch.mean
        for param_name, param_value in d.items():
            if isinstance(param_value, torch.Tensor):
                if param_name.endswith("_grad_change"):
                    buffer["grad_changes"].append(aggregator(param_value))
                elif param_name.endswith("_grad"):
                    buffer["grads"].append(aggregator(param_value))
                elif param_name.endswith("_change"):
                    buffer["param_changes"].append(aggregator(param_value))
                else:
                    buffer["params"].append(aggregator(param_value))
            else:
                res[param_name] = param_value

        assert not buffer["grads"] or len(buffer["params"]) == len(buffer["grads"])
        assert not buffer["param_changes"] or (len(buffer["param_changes"]) == len(buffer["params"]))

        for k, v in buffer.items():
            if v:
                res[f"{k}_aggregated"] = wandb.Histogram(v)  # type: ignore
    else:
        raise NotImplementedError()
    return res


def _extract_data_by_regex(runs: List[Dict], r, convert_to_int: Callable) -> List:
    data = []
    for run in runs:
        summary = run.summary
        summary_keys = list(summary.keys())
        # breakhis20/tl.bestval.breakhis20epoch.acc
        found = list(filter(r.match, summary_keys))
        d = [[convert_to_int(entry), summary[entry]] for entry in found]
        d.sort()
        for x, y in d:
            data.append([x, y])

    data.sort()
    return data


def _create_line_from_data(data: List, name: str, colors, confidence=False) -> List:
    figures = []
    if not confidence:
        x = [a[0] for a in data]
        y = [a[1] for a in data]
        figures.append(go.Scatter(x=x, y=y, name=name, fillcolor=colors[0]))
    if confidence:
        x = [a[0] for a in data]
        higher = []
        lower = []
        y = []
        unique, idxs, counts = np.unique(x, return_index=True, return_counts=True)
        for i, idx in enumerate(idxs):
            # if counts[i]>1:
            points = data[idx : idx + counts[i]]
            y_tmp = np.mean(points, axis=0)
            y.append(y_tmp[-1])
            higher.append(np.max(points, axis=0)[-1])
            lower.append(np.min(points, axis=0)[-1])
        unique = list(unique)
        higher = list(higher)
        lower = list(lower)
        y = list(y)
        figures.append(go.Scatter(x=unique, y=y, name=name, legendgroup=name, line_color=colors[0]))
        figures.append(
            go.Scatter(
                x=unique + unique[::-1],
                y=higher + lower[::-1],
                fill="toself",
                fillcolor=colors[1],
                line_color="rgba(255,255,255,0)",
                legendgroup=name,
                hoverinfo="skip",
            )
        )

    return figures


def create_sample_efficiency_plot(
    run_id: str,
    project_name: str,
    downstream_tasks,
    entity: str = "tissue-concepts",
    metric: str = "acc",
    compare_to_imageNet=None,
    compare_to_st=None,
    include_traditional_models=[],
):
    colors = px.colors.qualitative.Set2
    for color in px.colors.qualitative.Set1:
        colors.append(color)
    # colors = []
    # for color in px_colors:
    #    colors.append(f"rgba{color[3:-1]},0.3)")
    api = wandb.Api()
    run = api.run(f"{entity}/{project_name}/{run_id}/")

    fig = go.Figure()
    color_idx = 0

    # Evaluatr task performance
    for idx, task in enumerate(downstream_tasks):
        # clf head
        r = re.compile(f"{task}(.*)/tl.bestval.{task}(.*)epoch.{metric}")
        data = _extract_data_by_regex([run], r, lambda x: int(x.split("/")[0].split(f"{task}")[-1]))
        for img in _create_line_from_data(data, f"{task}", [colors[color_idx]]):
            fig.add_trace(img)
            color_idx += 1

        # img baseline
        if compare_to_imageNet is not None:
            r = re.compile(f"{task}(.*)/tl.bestval.{task}(.*).{metric}")
            data = _extract_data_by_regex(
                runs=api.runs(
                    path=compare_to_imageNet,
                    filters={"display_name": {"$regex": f"ImageNet-binary"}},
                ),
                r=r,
                convert_to_int=lambda x: int(x.split(f"{task}")[-1].split("_")[0].split("epoch")[0]),
            )
            for img in _create_line_from_data(
                data,
                f"{task}",
                [colors[color_idx], f"rgba{colors[color_idx][3:-1]},0.3)"],
                confidence=True,
            ):
                fig.add_trace(img)
                color_idx += 1

        if compare_to_st is not None:
            r = re.compile(f"{task}(.*)/tl.bestval.{task}(.*).{metric}")
            data = _extract_data_by_regex(
                runs=api.runs(
                    path=compare_to_imageNet,
                    filters={"display_name": {"$regex": f"{task}(.*)"}},
                ),
                r=r,
                convert_to_int=lambda x: int(x.split(f"{task}")[-1].split("_")[0].split("epoch")[0]),
            )
            for img in _create_line_from_data(
                data,
                f"{task}-ST",
                [colors[color_idx], f"rgba{colors[color_idx][3:-1]},0.3)"],
                confidence=True,
            ):
                fig.add_trace(img)
                color_idx += 1

        for model in include_traditional_models:
            r = re.compile(f"{task}(.*).{model}_val_{metric}")
            data = _extract_data_by_regex(
                runs=[run],
                r=r,
                convert_to_int=lambda x: int(x.split(f"{task}")[1].split("_")[0]),
            )
            for img in _create_line_from_data(data, f"{model}-{task}", [colors[color_idx]]):
                fig.add_trace(img)
                color_idx += 1

                if compare_to_imageNet is not None:
                    data = _extract_data_by_regex(
                        runs=api.runs(
                            path=compare_to_imageNet,
                            filters={"display_name": {"$regex": f"ImageNet-binary"}},
                        ),
                        r=r,
                        convert_to_int=lambda x: int(x.split(f"epoch")[0].split(f"{task}")[-1].split("_")[0]),
                    )
                    for img in _create_line_from_data(
                        data,
                        f"{model}-ImageNet",
                        [colors[color_idx], f"rgba{colors[color_idx][3:-1]},0.3)"],
                        confidence=True,
                    ):
                        fig.add_trace(img)
                        color_idx += 1

    fig.update_layout(xaxis_title="Percent Samples", yaxis_title=f"best overall {metric}")

    return fig


def _get_imgnet_defaults(api, path, task, metric):
    runs = api.runs(path=path, filters={"display_name": {"$regex": f"ImageNet"}})
    data = []
    for run in runs:
        summary = run.summary
        summary_keys = list(summary.keys())
        r = re.compile(f"epoch_tl.bestval(.*){task}(.*)_{metric}")
        found = list(filter(r.match, summary_keys))
        d = [[int(entry.split(f"{task}")[-1].split("_")[0]), summary[entry]] for entry in found]

        d.sort()
        for x, y in d:
            data.append([x, y])

    data.sort()
    x = [a[0] for a in data]
    higher = []
    lower = []
    y = []
    unique, idxs, counts = np.unique(x, return_index=True, return_counts=True)
    for i, idx in enumerate(idxs):
        points = data[idx : idx + counts[i]]
        y_tmp = np.mean(points, axis=0)
        y.append(y_tmp[-1])
        higher.append(np.max(points, axis=0)[-1])
        lower.append(np.min(points, axis=0)[-1])
    unique = list(unique)
    higher = list(higher)
    lower = list(lower)
    y = list(y)

    line = go.Scatter(x=unique, y=y, name=f"ImageNet-{task}")

    return [
        go.Scatter(
            x=unique + unique[::-1],
            y=higher + lower[::-1],
            fill="toself",
            fillcolor="rgba(205,205,205,0.5)",
            line_color="rgba(255,255,255,0)",
            showlegend=False,
            hoverinfo="skip",
            name=f"ImageNet-{task}",
        ),
        line,
    ]


def _get_st_plot(api, path, task, metric):
    runs = api.runs(path=path, filters={"display_name": {"$regex": f"{task}[0-9][0-9][0-9]"}})
    data = []
    for run in runs:
        name = run.name
        last = run.summary[f"epoch_mtlval_{metric}"]
        perc = int(name.split(f"{task}")[-1].split("_")[0])
        data.append([perc, last])

    data.sort()
    x = [a[0] for a in data]
    # y = [a[1] for a in data]

    higher = []
    lower = []
    y = []
    unique, idxs, counts = np.unique(x, return_index=True, return_counts=True)
    for i, idx in enumerate(idxs):
        # if counts[i]>1:
        points = data[idx : idx + counts[i]]
        y_tmp = np.mean(points, axis=0)
        y.append(y_tmp[-1])
        higher.append(np.max(points, axis=0)[-1])
        lower.append(np.min(points, axis=0)[-1])
    unique = list(unique)
    higher = list(higher)
    lower = list(lower)
    y = list(y)

    line = go.Scatter(x=unique, y=y, name=f"SingleTask-{task}")

    return [
        go.Scatter(
            x=unique + unique[::-1],
            y=higher + lower[::-1],
            fill="toself",
            fillcolor="rgba(205,205,205,0.5)",
            line_color="rgba(255,255,255,0)",
            showlegend=False,
            hoverinfo="skip",
            name=f"SingleTask-{task}",
        ),
        line,
    ]


@torch.no_grad()
def multitask_embedding_table(ts: List[MTLTask], enc: PyramidEncoder, max_batches: int) -> wandb.Table:
    from mmm.task_sampling import BalancedTaskSampler
    from tqdm.auto import tqdm

    logtable = wandb.Table(columns=["taskid", "tooltip", "repr", "internal_group"])

    for t in ts:
        t.cohort.prepare_epoch(epoch=0)

    tasksampler = BalancedTaskSampler(BalancedTaskSampler.Config(), ts, DataSplit.val)

    for batchcounter, (batch, task) in tqdm(enumerate(tasksampler)):
        _, embs = enc(batch["image"].to(enc.torch_device))
        cls_names: List[str] = task.class_names  # type: ignore
        for batch_idx in range(len(batch)):
            class_name = cls_names[batch["class"][batch_idx].item()]
            img = wandb.Image(batch["image"][batch_idx, ...].detach().cpu(), caption=f"{class_name}")
            logtable.add_data(
                task.get_name(),
                img,
                list(embs[batch_idx].detach().cpu().numpy()),
                f"{task.get_name()}_{class_name}",
            )

        if batchcounter >= max_batches:
            break

    return logtable



from __future__ import annotations
from typing import Any, Dict, Tuple

import torchvision.transforms.functional as F

from torch.utils.data import Dataset

from mmm.logging.st_ext import blend_with_mask
from .MTLDataset import MTLDataset, SrcCaseType


class RegressionDataset(MTLDataset):
    def __init__(self, src_ds: Dataset[SrcCaseType], *args, **kwargs) -> None:
        super().__init__(src_ds, ["image", "target"], ["meta"], *args, **kwargs)

    def verify_case_by_index(self, index: int) -> Dict[str, Any]:
        case = super().verify_case_by_index(index)
        self.assert_image_data_assumptions(case["image"])
        return case

    def get_input_output_tuple(self, batch: Dict[str, Any]) -> Tuple[Any, ...]:
        return batch["image"], batch["target"]

    def st_case_viewer(self, case: Dict[str, Any], index: int = -1) -> None:
        import streamlit as st

        st.write(f"Target: {case['target']}")
        im = case["image"]
        blend_with_mask(im, None, caption_suffix=f"Shape: {im.shape}", st_key=f"c{index}")
        st.write(case)

    def _compute_batchsize_from_batch(self, batch: Dict[str, Any]) -> int:
        return batch["image"].shape[0]

    def _visualize_batch_case(self, batch: Dict[str, Any], i: int) -> None:
        import streamlit as st

        patch = batch["image"][i]
        st.write(f"Target: {batch['target'][i]}")
        blend_with_mask(
            patch,
            None,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}: {patch.shape}",
            st_key=f"b{i}",
        )



from __future__ import annotations
from pathlib import Path
import logging
import itertools
from typing_extensions import Annotated
from pydantic import Field, field_validator, model_validator
from typing import Literal
import numpy as np
from tiffslide import TiffSlide
import cv2
import torch
from tqdm.auto import tqdm
from torch.utils.data import Dataset, DataLoader
import torchvision.transforms.functional as F
from torchvision.transforms import InterpolationMode
from torchvision.transforms._functional_tensor import _get_gaussian_kernel2d

try:
    from fastapi import APIRouter
except ImportError:
    APIRouter = None

from mmm.labelstudio_ext.projects import LSProject
from mmm.labelstudio_ext.LSModel import LSModel, LabelStudioTask, SerializedArray
from mmm.labelstudio_ext.utils import download_image
from mmm.labelstudio_ext.utils import binary_mask_to_result
from mmm.labelstudio_ext.NativeBlocks import NativeBlocks
from mmm.data_loading.DistributedPath import DistributedPath
from mmm.transforms import UnifySizes
from mmm.BaseModel import BaseModel
from mmm.mtl_modules.tasks.SemSegTask import SemSegTask
from mmm.data_loading.geojson.utils import (
    annotations_from_mask,
    create_featurecollection,
)


def compute_uncertainty_score(probas, preds):
    # For multiclass segmentation, probas has shape [C, H, W] between 0 and one
    uncertainy = torch.stack([torch.min(probas[i], 1 - probas[i]) * 2 for i in range(len(probas))])  # each class
    mean_per_class = torch.mean(uncertainy, dim=(1, 2))
    return torch.mean(mean_per_class).item()
    # return torch.mean(uncertainy).item()


def by_number_of_blobs(probas, preds):
    def count_blobs_in_mask(mask):
        # Each mask is (H, W)
        return len(
            cv2.findContours(
                mask.cpu().numpy().astype("uint8"),
                cv2.RETR_EXTERNAL,
                cv2.CHAIN_APPROX_SIMPLE,
            )[0]
        )

    if len(preds.shape) == 3:
        # multi-label
        num_blobs_by_class = [count_blobs_in_mask(preds[i]) for i in range(len(preds))]
    else:
        # multi-class (H, W)
        num_blobs_by_class = [count_blobs_in_mask(preds)]
    return sum(num_blobs_by_class)


class SingleInvocation(BaseModel):
    invocation_type: Literal["single"] = "single"
    task: LabelStudioTask
    mtl_task_id: str
    return_type: Literal["predictions", "probabilities"] = "predictions"


class WSIWindows(Dataset):
    def __init__(
        self,
        coords: list[tuple[int, int]],
        window_size: int,
        slide: TiffSlide,
        level: int,
    ) -> None:
        self.coords, self.window_size, self.slide, self.level = (
            coords,
            window_size,
            slide,
            level,
        )

    def __len__(self):
        return len(self.coords)

    def __getitem__(self, index) -> torch.Tensor:
        row, col = self.coords[index]
        downsample_fac = self.slide.level_downsamples[self.level]
        row_l0, col_l0 = row * downsample_fac, col * downsample_fac
        try:
            img = self.slide.read_region(
                (col_l0, row_l0),
                self.level,
                (self.window_size, self.window_size),
                as_array=True,
            )
        except Exception as e:
            img = np.zeros((self.window_size, self.window_size, 3), dtype=np.uint8)
            logging.error(f"Error while reading region {(row, col)}: {e}")
        return {"image": F.to_tensor(img), "coords": (row, col), "level": self.level}


class WSIInvocation(BaseModel):
    """
    GeoJSON inference for WSI.
    """

    invocation_type: Literal["wsi"] = "wsi"
    wsi_path: DistributedPath
    mtl_task_id: str
    for_levels: list[int] = [0]
    out_downsample_fac: int = Field(
        default=8,
        description="Downsample factor of the output mask, greatly helps with reducing runtime and network load",
    )
    patch_size: int = 1024
    batch_size: int = 4
    filter_tiny_obj: bool = False
    obj_min_area: float = Field(default=0.0, description="Minimum area of an object in pixels in level 0")
    minimum_confidence: float = Field(
        default=0.0,
        description="Minimum pixel confidence for considering a pixel as a class",
    )
    vote_type: Literal["uniform", "gaussian"] = "gaussian"

    @field_validator("for_levels")
    def check_for_levels(cls, v):
        if len(v) == 0:
            raise ValueError("for_levels must not be empty")
        if min(v) < 0:
            raise ValueError("for_levels must be >= 0")
        if len(set(v)) != len(v):
            raise ValueError("for_levels must be unique")
        return v

    def get_patch_coords(self, height: int, width: int) -> list[tuple[int, int]]:
        min_row, min_col = -self.patch_size // 2, -self.patch_size // 2
        max_row, max_col = height - self.patch_size // 2, width - self.patch_size // 2
        return [
            (row, col)
            for (row, col) in itertools.product(
                range(min_row, max_row + 1 + self.patch_size // 2, self.patch_size),
                range(min_col, max_col + 1 + self.patch_size // 2, self.patch_size),
            )
        ]

    async def wsi_inference(self, torch_modules: NativeBlocks):
        mtl_task: SemSegTask = torch_modules[self.mtl_task_id]
        slide = TiffSlide(self.wsi_path.file().open())

        assert len(self.for_levels) == 1, "For now, only single-level inference is implemented"

        # The lowest level determines the granularity in the stitched output
        lowest_level: int = min(self.for_levels)

        # Patchify exhaustively at the given levels
        patch_coords = self.get_patch_coords(*slide.level_dimensions[lowest_level][::-1])

        # Stitch together the probabilities of each class
        # by voting with gaussian masks such that the patch-borders have less impact than the center
        out_HW = tuple(
            map(
                lambda x: x // self.out_downsample_fac,
                slide.level_dimensions[lowest_level][::-1],
            )
        )
        stitch_CHW: torch.Tensor = torch.zeros((len(mtl_task.class_names),) + out_HW, dtype=torch.float32).to(
            torch_modules.get_device()
        )
        if self.vote_type == "uniform":
            vote_mask = torch.ones(
                (
                    self.batch_size,
                    len(mtl_task.class_names),
                    self.patch_size // self.out_downsample_fac,
                    self.patch_size // self.out_downsample_fac,
                ),
                dtype=torch.float32,
            ).to(torch_modules.get_device())
        elif self.vote_type == "gaussian":
            sigma = self.patch_size // 16
            single_vote = _get_gaussian_kernel2d(
                (
                    self.patch_size // self.out_downsample_fac,
                    self.patch_size // self.out_downsample_fac,
                ),
                [sigma, sigma],
                torch.float32,
                stitch_CHW.device,
            )
            # Scale such that the average vote is 1
            single_vote /= single_vote.mean()

            # Visualize heatmap and save
            # import matplotlib.pyplot as plt
            # import seaborn as sns
            # sns.heatmap(single_vote.cpu().numpy())
            # plt.savefig("heatmap.png")

            vote_mask = single_vote.repeat(self.batch_size, len(mtl_task.class_names), 1, 1)
        else:
            raise NotImplementedError()

        with torch.no_grad():
            # Organize into batches
            wsi_windows = WSIWindows(patch_coords, self.patch_size, slide, lowest_level)
            logging.info(f"WSI inference: {len(wsi_windows)} patches to build a {stitch_CHW.shape} mask")
            for batch in tqdm(DataLoader(wsi_windows, batch_size=self.batch_size, num_workers=0)):
                torch_input = batch["image"].to(torch_modules.get_device())
                logits = mtl_task.forward(torch_input, torch_modules)
                # probas = mtl_task.logits_to_probas(logits=logits, output_size=torch_input.shape[-2:])
                logits_resized = F.resize(
                    logits,
                    [
                        self.patch_size // self.out_downsample_fac,
                        self.patch_size // self.out_downsample_fac,
                    ],
                    interpolation=InterpolationMode.NEAREST_EXACT,
                )
                # Apply voting mask (H, W) to the last two dimensions of logits (B, C, H, W)
                # The last batch might be smaller than the batch size
                vote = logits_resized * vote_mask[: logits_resized.shape[0]]

                # Stamp the votes onto the stitched image
                for i in range(vote.shape[0]):
                    row, col = (
                        batch["coords"][0][i].item(),
                        batch["coords"][1][i].item(),
                    )
                    row, col = (
                        row // self.out_downsample_fac,
                        col // self.out_downsample_fac,
                    )
                    # Only use the valid region of the vote
                    vote_row_offset = max(0, -row)
                    vote_col_offset = max(0, -col)
                    vote_row_end = min(vote.shape[2], stitch_CHW.shape[1] - row)
                    vote_col_end = min(vote.shape[3], stitch_CHW.shape[2] - col)
                    vote_valid = vote[i][:, vote_row_offset:vote_row_end, vote_col_offset:vote_col_end]
                    stitch_CHW[
                        :,
                        row + vote_row_offset : row + vote_row_end,
                        col + vote_col_offset : col + vote_col_end,
                    ] += vote_valid

            # Normalize the stitched image
            stitchprobas = mtl_task.logits_to_probas(stitch_CHW.unsqueeze(0))
            stitchpreds = (
                mtl_task.probas_to_preds(stitchprobas, pixel_threshold=self.minimum_confidence).squeeze(0).cpu().numpy()
            )

        regions = annotations_from_mask(
            stitchpreds,
            for_values={
                class_index: class_name for class_index, class_name in enumerate(mtl_task.class_names)
            },  # {1: "tumor", 2: "stroma", 3: "necrosis"},
            downsample_fac=slide.level_downsamples[lowest_level] * self.out_downsample_fac,
            coarse=self.filter_tiny_obj,
            min_area=self.obj_min_area,
        )
        outgeojson_dict: dict = create_featurecollection(regions)

        return outgeojson_dict


class ModelInput(BaseModel):
    """
    This model itself is not a union because FastAPI doesn't like it.
    On their main branch it is already fixed by the time of writing.
    """

    invo: Annotated[SingleInvocation | WSIInvocation, Field(discriminator="invocation_type")]


class DLModel(LSModel):
    """
    A deep learning model does not use the train route.
    """

    class Config(LSModel.Config):
        category: str = "DeepLearning"
        pixel_confidence_threshold: float = 0.5
        region_confidence_threshold: float = 0.0

    def __init__(self, cfg: Config, ls_client, prefix: str, blocks: NativeBlocks) -> None:
        self.cfg: DLModel.Config
        self.model_version = "t24"
        self.torch_modules = blocks
        self.shared_blocks: list[str] = self.torch_modules.get_sharedblock_keys()
        self.tasks: list[str] = self.torch_modules.get_task_keys()
        super().__init__(cfg, ls_client, prefix)

    @torch.no_grad()
    async def predict_for_tensor(self, mtl_task: SemSegTask, torch_input: torch.Tensor):
        logits = mtl_task.forward(torch_input, self.torch_modules)
        probas = mtl_task.logits_to_probas(logits=logits, output_size=torch_input.shape[-2:])
        preds = mtl_task.probas_to_preds(probas, pixel_threshold=self.cfg.pixel_confidence_threshold)
        return probas, preds

    @torch.no_grad()
    async def predict_for_task(self, mtl_task, ls_task: dict):
        input_image = download_image(ls_task["data"]["image"], ls_client=self.ls_client)
        torch_input = (
            UnifySizes(divisable_by=32, enforce_order=True)([{"image": F.to_tensor(input_image)}])[0]["image"]
            .unsqueeze(0)
            .to(self.torch_modules.get_device())
        )
        probas, preds = await self.predict_for_tensor(mtl_task, torch_input)
        return probas.squeeze(0), preds.squeeze(0)

    @torch.no_grad()
    async def predict_single_mcseg(self, ls_task: dict, task_id: str, label, cfg):
        mtl_task: SemSegTask = self.torch_modules[task_id]
        probas, preds = await self.predict_for_task(mtl_task, ls_task)
        all_results, all_scores = [], []
        for i in range(len(mtl_task.class_names)):
            score = probas[i][preds == i].mean().item()
            res = binary_mask_to_result(
                (preds == i).cpu().numpy(),
                mtl_task.class_names[i],
                mtl_task.get_name(),
                score=score,
            )
            if score > self.cfg.region_confidence_threshold:
                all_results.append(res)
                all_scores.append(score)

        # result_dict = {"result": , "score": sum(all_scores) / len(all_scores) if len(all_scores) > 0 else 0}
        # result_dict.update(await self.get_version())
        return all_results, all_scores

    @torch.no_grad()
    async def predict_single(self, project: LSProject, ls_task: dict, context: dict | None = None):
        good_results, prediction_scores = [], []
        for label, cfg in project.ls.parsed_label_config.items():
            matching_tasks = [taskid for taskid in self.tasks if label.startswith(taskid)]
            for taskid in matching_tasks:
                results, scores = await self.predict_single_mcseg(ls_task, taskid, label, cfg)
                good_results.extend(results)
                prediction_scores.extend(scores)
        return {
            "result": good_results,
            "model_version": self.model_version,
            "score": np.mean(prediction_scores) if len(prediction_scores) > 0 else 0,
        }

    @torch.no_grad()
    async def predict(self, project: LSProject, tasks: list[dict], context: dict | None = None):
        if len(tasks) >= 1:
            return {"results": [await self.predict_single(project, lstask, context) for lstask in tasks]}
        return {}

    async def labeling_priority(
        self,
        task_id: str,
        ls_task: LabelStudioTask,
        scoring_type: Literal["uncertainty", "by_number_of_blobs"] = "uncertainty",
    ) -> float:
        if task_id not in self.tasks:
            return f"Task {task_id} not found"
        mtl_task = self.torch_modules[task_id]
        probas, preds = await self.predict_for_task(mtl_task, ls_task.model_dump())
        # Render tensor as json
        # return {"probas": probas.cpu().numpy().tolist(), "preds": preds.cpu().numpy().tolist()}
        scorers = {
            "uncertainty": compute_uncertainty_score,
            "by_number_of_blobs": by_number_of_blobs,
        }
        return scorers[scoring_type](probas, preds)

    async def invocation(self, inps: ModelInput) -> SerializedArray | dict:
        """
        Invokes the model on a single input and returns the raw result.
        """
        invo = inps.invo
        if invo.invocation_type == "single":
            if invo.mtl_task_id not in self.tasks:
                return f"Task {invo.mtl_task_id} not found"
            mtl_task = self.torch_modules[invo.mtl_task_id]
            probas, preds = await self.predict_for_task(mtl_task, invo.ls_task.model_dump())
            if invo.return_type == "probabilities":
                return SerializedArray.from_numpy(probas.cpu().numpy())
            elif invo.return_type == "predictions":
                return SerializedArray.from_numpy(preds.cpu().numpy())
            else:
                raise NotImplementedError()
        elif invo.invocation_type == "wsi":
            geojson = await invo.wsi_inference(self.torch_modules)
            return geojson

    def get_task_ids(self) -> list[str]:
        return self.tasks

    def get_shared_block_ids(self) -> list[str]:
        return self.shared_blocks

    def get_task_info(self, task_id: str) -> dict:
        res = self.torch_modules[task_id].args.model_dump()
        for attr in ["class_names", "training_state", "torch_device"]:
            if hasattr(self.torch_modules[task_id], attr):
                res[attr] = getattr(self.torch_modules[task_id], attr)
        return res

    def build_router(self) -> APIRouter:
        res = super().build_router()
        res.get("/task_ids", tags=[self.get_tag()])(self.get_task_ids)
        res.get("/task_info/{task_id}", tags=[self.get_tag()])(self.get_task_info)
        res.get("/shared_block_ids", tags=[self.get_tag()])(self.get_shared_block_ids)
        return res

    def get_versions(self):
        return {"versions": [self.model_version]}



from __future__ import annotations
from typing import Literal, Any
import jinja2

import numpy as np
import torch
from torchvision.transforms._functional_tensor import _get_gaussian_kernel2d
import uuid
import cv2
import torch.nn as nn
import torch.nn.functional as nnF
import logging
from pydantic import Field

try:
    import lightgbm as lgb
    from label_studio_sdk import Project
except ImportError:
    logging.warning("api extra missing")
    lgb, Project = None, None

from mmm.labelstudio_ext.projects import LSProject
from mmm.labelstudio_ext.LSModel import LSModel, Project
from mmm.labelstudio_ext.LSDataset import LSDataset
from mmm.labelstudio_ext.utils import (
    download_image,
    convert_task_to_seglabel,
    brush_annotation_to_npy,
    binary_mask_to_result,
)
from mmm.transforms import UnifySizes
from mmm.data_loading.MultiLabelSemSegDataset import masks_from_ls
from mmm.BaseModel import BaseModel
from mmm.labelstudio_ext.NativeBlocks import NativeBlocks
import torchvision.transforms.functional as F
from sklearn.model_selection import train_test_split


def compute_pixel_features_for_sklearn(feature_maps, keep_batch_dim: bool, masks: torch.Tensor, ignore_index: int = 0):
    """Computation of the dense features

    Take for example these feature maps from the encoder and decoder for one image:

    ```python
    feature_maps = torch.Tensor([
        [[0, 0], [2, 0]],
        [[1, 0], [1, 1]],
        [[2, 4], [2, 2]],
    ])
    feature_maps = torch.stack([feature_maps.clone(), (feature_maps + 3).clone()]) # (B=2, C=3, H=2, W=2)
    # or:
    feature_maps = torch.stack([
        torch.stack([torch.arange(16).view(4, 4) for i in range(3)])
        for _ in range(2)])
    masks = torch.stack([torch.arange(16).view(4, 4) for _ in range(2)])
    ```
    The first feature map is [[0, 0], [0, 0]] and the last is [[2, 2], [2, 2]].
    The first pixel should have the feature vector [0, 1, 2].
    The pixel features should have shape (8, 3) or (B=2, 4, C=3)
    """
    assert feature_maps.shape[0] == masks.shape[0], f"{feature_maps.shape=} {masks.shape=}"

    # Transpose the channel dimension to the end
    transposed = feature_maps.permute(0, 2, 3, 1)
    if keep_batch_dim:
        # Reshape into (B, H*W, C)
        flattened = transposed.contiguous().view(transposed.size(0), -1, transposed.size(3))
        flat_mask = masks.contiguous().view(masks.size(0), -1)
    else:
        # Reshape into (H*W, C)
        flattened = transposed.contiguous().view(-1, transposed.size(3))
        flat_mask = masks.contiguous().view(-1)

    # Return only the pixels that are not ignored
    return flattened[flat_mask != ignore_index], flat_mask[flat_mask != ignore_index]


class LGBBase:
    async def update_model(self, features: tuple, label, cfg, tasks):
        raise NotImplementedError

    async def predict_results(self, features, label, cfg, project: LSProject, task: dict, context: dict | None):
        raise NotImplementedError


class LGBMCSegmentor(LGBBase):
    def __init__(self, model_cfg: LGBModel.Config) -> None:
        self.class_names, self.model = None, None
        self.model_cfg = model_cfg
        self.X, self.y = None, None

    async def update_model(self, features: tuple, label, cfg, tasks):
        dense_features = features["image"][0]
        # Compute label maps from RLE representation in tasks
        self.class_names = cfg["labels"]
        full_masks = [convert_task_to_seglabel(t, ["Unsure"] + self.class_names) for t in tasks]
        # Resize all masks to be the same size as the features
        resized_masks = [
            (
                F.resize(
                    torch.Tensor(mask).unsqueeze(0),
                    feat.shape[-2:],
                    interpolation=F.InterpolationMode.NEAREST,
                )[0]
                if mask is not None
                else None
            )
            for feat, mask in zip(dense_features, full_masks)
        ]

        # Filter all pixels that have no label
        Xy = [
            compute_pixel_features_for_sklearn(
                dense_features[i],
                keep_batch_dim=False,
                masks=resized_masks[i].unsqueeze(0),
            )
            for i in range(len(tasks))
            if resized_masks[i] is not None
        ]
        X, y = torch.cat([x[0] for x in Xy]), torch.cat([x[1] for x in Xy])
        X_train, X_val, y_train, y_val = train_test_split(X.numpy(), y.numpy() - 1, test_size=0.2)

        # If only a single task is given, we add it to the training data
        if self.X is None:
            self.X, self.y = (X_train, X_val), (y_train, y_val)
        else:
            self.X = (
                np.concatenate([self.X[0], X_train]),
                np.concatenate([self.X[1], X_val]),
            )
            self.y = (
                np.concatenate([self.y[0], y_train]),
                np.concatenate([self.y[1], y_val]),
            )
            # self.X, self.y = torch.cat([self.X, X]), torch.cat([self.y, y])
        logging.info(f"Training segmentor for {label} with {self.X[0].shape=}, {self.y[0].shape=}")
        train_ds = lgb.Dataset(self.X[0], label=self.y[0])
        val_ds = lgb.Dataset(self.X[1], label=self.y[1], reference=train_ds)

        if self.model is None:
            self.model = lgb.train(
                {
                    "objective": "multiclass",
                    "num_class": len(self.class_names),
                    "force_col_wise": True,
                    "metric": {"multi_logloss"},
                },  # , "device_type": "cuda"},
                train_ds,
                # keep_training_booster=True,
                num_boost_round=100,
                valid_sets=[val_ds],
                callbacks=[
                    lgb.early_stopping(stopping_rounds=3),
                ],
            )
        else:
            logging.info(f"{self.model=} already exists, only train on the new data")
            self.model = lgb.train(
                {
                    "objective": "multiclass",
                    "num_class": len(self.class_names),
                    "force_col_wise": True,
                    "metric": {"multi_logloss"},
                },  # , "device_type": "cuda"},
                train_ds,
                init_model=self.model,
                num_boost_round=10,
                valid_sets=[val_ds],
                callbacks=[
                    lgb.early_stopping(stopping_rounds=3),
                ],
            )

        # self.model = lgb.train(
        #     {
        #         "objective": "multiclass",
        #         "num_class": len(self.class_names),
        #         "boosting_type": "rf",
        #         "num_trees": 100,
        #         "bagging_fraction": 0.9,  # use fraction of the data for each tree
        #         "feature_fraction": 0.9,  # use fraction of the features for each tree
        #     },
        #     lgb.Dataset(X.numpy(), label=y.numpy() - 1),
        # )
        logging.info(f"Trained segmentor {self} for {label} with {self.X[0].shape=}, {self.y[0].shape=}")
        return self

    async def predict_results(self, features, label, cfg, project: LSProject, task: dict, context: dict | None):
        if self.model is None:
            logging.warning(f"Skipping {label} because the model is not trained yet")
            return [], []

        spatial_features = features[cfg["inputs"][0]["value"]][0][0]
        img_size = features[cfg["inputs"][0]["value"]][-1][0]
        X = compute_pixel_features_for_sklearn(
            spatial_features,
            keep_batch_dim=False,
            masks=torch.ones(spatial_features.shape[0], *spatial_features.shape[-2:]),
        )[0].numpy()
        probas = torch.Tensor(self.model.predict(X))
        probas_spatial = probas.view(*spatial_features.shape[-2:], len(self.class_names))
        probas_spatial = F.resize(
            probas_spatial.permute(2, 0, 1),
            img_size,
            interpolation=F.InterpolationMode.NEAREST_EXACT,
        ).permute(1, 2, 0)
        y_hat = torch.argmax(probas_spatial, axis=-1)
        # Set those pixels to -1 that have a confidence lower than the threshold
        y_hat[probas_spatial.max(axis=-1).values < self.model_cfg.pixel_confidence_threshold] = -1

        # y_hat_original = F.resize(
        #     y_hat.unsqueeze(0),
        #     [x * 2 for x in y_hat.shape],
        #     # download_image(task["data"][cfg["inputs"][0]["value"]]).shape[:2],
        #     interpolation=F.InterpolationMode.NEAREST,
        # )[0]

        results, prediction_scores = [], []
        for i, class_name in enumerate(self.class_names):
            if self.model_cfg.split_connected_comp:
                num_labels, labels, stats, _ = cv2.connectedComponentsWithStats((y_hat == i).numpy().astype(np.uint8))
                areas = [s[4] for s in stats]
            else:
                # The first label is the background
                num_labels = 2 if i in y_hat else 0
                labels = (y_hat == i).numpy().astype(np.uint8)
                # count the ones in the mask
                areas = [None, torch.sum(y_hat == i).item()]

            for label_index in range(1, num_labels):
                region_mask = labels == label_index
                region_score = torch.mean(probas_spatial[region_mask][:, i]).item()
                if (
                    region_score > self.model_cfg.region_confidence_threshold
                    and areas[label_index] > self.model_cfg.region_minimum_area
                ):
                    results.append(binary_mask_to_result(region_mask, class_name, label, score=region_score))
                    prediction_scores.append(region_score)

        # prediction_scores = [torch.mean(probas_spatial[:, :, i]).item() for i in range(len(self.class_names))]
        # Only take the average across the region of that class
        # prediction_scores = [
        #     torch.mean(probas_spatial[y_hat == i][:, i]).item() if i in y_hat else 0
        #     for i in range(len(self.class_names))
        # ]

        # Scale back from inference size to original image size

        # results = [
        #     binary_mask_to_result((y_hat_original == i).numpy(), self.class_names[i], label, score=prediction_scores[i])
        #     for i in range(len(self.class_names))
        #     if i in y_hat_original
        # ]
        return results, prediction_scores


class LGBMLSegmentor(LGBBase):
    def __init__(self, model_cfg: LGBModel.Config) -> None:
        self.class_names, self.models = None, None
        self.model_cfg = model_cfg

    async def update_model(self, features: tuple, label, cfg, tasks):
        dense_features = features[cfg["inputs"][0]["value"]][0]
        # Compute label maps from RLE representation in tasks
        self.class_names = cfg["labels"]

        def get_results(task):
            return [
                result
                for annotation in task["annotations"]
                for result in annotation["result"]
                if result["type"] == "brushlabels"
            ]

        full_masks = [torch.Tensor(masks_from_ls(get_results(t), None, self.class_names)) for t in tasks]
        # Resize all masks to be the same size as the features
        resized_masks = [
            F.resize(
                mask,
                dense_features.shape[-2:],
                interpolation=F.InterpolationMode.NEAREST,
            )
            for mask in full_masks
        ]

        # For each class, train a binary classifier
        self.models = {}
        for i, class_name in enumerate(self.class_names):
            # Filter all pixels that have no label
            X, y = compute_pixel_features_for_sklearn(
                dense_features,
                keep_batch_dim=False,
                masks=torch.stack([m[i] for m in resized_masks]),
                ignore_index=-1,
            )
            logging.info(f"Training segmentor for {label} with {X.shape=}, {y.shape=}")
            self.models[class_name] = lgb.train(
                {"objective": "binary"},
                lgb.Dataset(X.numpy(), label=y.numpy()),
            )
            logging.info(f"Trained segmentor {self} for {label} with {X.shape=}, {y.shape=}")
        return self

    async def predict_results(self, features, label, cfg, project: LSProject, task: dict, context: dict | None):
        if self.models is None:
            logging.warning(f"Skipping {label} because the model is not trained yet")
            return [], []

        spatial_features = features[cfg["inputs"][0]["value"]][0]
        X = compute_pixel_features_for_sklearn(
            spatial_features,
            keep_batch_dim=False,
            masks=torch.ones(spatial_features.shape[0], *spatial_features.shape[-2:]),
        )[0].numpy()

        results, prediction_scores = [], []
        for i, class_name in enumerate(self.class_names):
            probas = torch.Tensor(self.models[class_name].predict(X))
            # H, W
            probas_spatial = probas.view(*spatial_features.shape[-2:])
            probas_spatial = F.resize(
                probas_spatial.unsqueeze(0),
                [x * 2 for x in probas_spatial.shape],
                # download_image(task["data"][cfg["inputs"][0]["value"]]).shape[:2],
                interpolation=F.InterpolationMode.NEAREST_EXACT,
            )[0]
            y_hat = probas_spatial > self.model_cfg.pixel_confidence_threshold

            # Process each connected component separately
            if self.model_cfg.split_connected_comp:
                num_labels, labels, stats, _ = cv2.connectedComponentsWithStats(y_hat.numpy().astype(np.uint8))
                areas = [s[4] for s in stats]
            else:
                num_labels = 2
                labels = y_hat.numpy().astype(np.uint8)
                # count the ones in the mask
                areas = [None, torch.sum(y_hat).item()]

            # num_labels, labels = cv2.connectedComponents(y_hat.numpy().astype(np.uint8))
            for label_index in range(1, num_labels):
                # left, top, width, height, area =
                # area =
                component_mask = labels == label_index
                region_score = torch.mean(probas_spatial[component_mask]).item()

                # Filter regions
                if (
                    region_score > self.model_cfg.region_confidence_threshold
                    and areas[label_index] > self.model_cfg.region_minimum_area
                ):
                    results.append(binary_mask_to_result(component_mask, class_name, label, score=region_score))
                    prediction_scores.append(region_score)
        return results, prediction_scores


class LGBModel(LSModel):
    """
    A model based on lightgbm.

    Initializes from an LSDataset for loading the annotations, and respects the train route for updating the model.
    """

    class Config(LSModel.Config):
        category: str = "LightGBM"
        # trainer_checkpoint: str = Field(description="The mtl trainer checkpoint is used for extracting features.")

        split_connected_comp: bool = Field(
            default=False,
            description="All class names that match will be processed using connected components.",
        )
        pixel_confidence_threshold: float = Field(
            default=0.5,
            description="Only pixels with a confidence greater than this threshold will be considered.",
        )
        region_confidence_threshold: float = Field(
            default=0.8,
            description="Only regions with a confidence greater than this threshold will be considered.",
        )
        region_minimum_area: int = Field(
            default=100,
            description="Only regions with an area greater than this threshold will be considered.",
        )

        neighbourhood_features: Literal["none", "avgpool", "blur"] = "avgpool"

        segmentation_mode: Literal["multiclass", "multilabel"] = "multiclass"

    def __init__(self, cfg: Config, ls_client, prefix: str, torch_modules: NativeBlocks) -> None:
        self.cfg: LGBModel.Config
        self.torch_modules = torch_modules
        self.pooling: nn.Module = nn.AdaptiveAvgPool2d((1, 1))
        self.flatten = nn.Flatten(1)
        self.version = 0
        self.env = jinja2.Environment(loader=jinja2.PackageLoader("mmm", "resources"))

        # Feature buffer for each (project, tag). Usually just one called "image".
        self._feature_buffer: dict[str, tuple] = {}

        # For each Choices label there should be a trained classifier
        self.classifiers: dict[tuple[int, str], Any] = {}
        self.segmentors: dict[tuple[int, str], Any] = {}

        super().__init__(cfg, ls_client, prefix)

    async def get_user_html(self) -> str:
        return self.env.get_template("lgbmodel_status.jinja2").render(
            {
                "prefix": self.prefix,
                "config": self.cfg,
                "model": self,
                "classifiers": self.classifiers,
                "segmentors": self.segmentors,
            }
        )

    @torch.no_grad()
    async def predict_single_clf(self, features, label, cfg, project: LSProject, task: dict, context: dict | None):
        res, scores = [], []

        model, class_names = self.classifiers[label]
        X = features[cfg["inputs"][0]["value"]][1].numpy()
        logging.info(f"Predicting {label} for {X.shape=}")
        y_hat = model.predict(X)[0]
        best_index = np.argmax(y_hat)
        res.append(
            {
                "id": str(uuid.uuid4())[0:8],
                "type": "choices",
                "value": {"choices": [class_names[best_index]]},
                "origin": "manual",
                "to_name": cfg["inputs"][0]["value"],
                "from_name": label,
                "image_rotation": 0,
                "score": y_hat[best_index],
            }
        )
        scores.append(y_hat[best_index])
        logging.info([f"{name}: {x:.3f}" for name, x in zip(class_names, y_hat)])

        return res, scores

    @torch.no_grad()
    async def predict_single(self, project: LSProject, task: dict, context: dict | None):
        goodresults = []
        prediction_scores = []
        features = {}
        for label, cfg in project.ls.parsed_label_config.items():
            for k in [inp["value"] for inp in cfg["inputs"] if inp["type"] == "Image"]:
                if k not in features:
                    features[k] = self.get_features([task], k)
            if (model_key := (project.ls.id, label)) in self.classifiers:
                results, scores = await self.predict_single_clf(features, label, cfg, project, task, context)
                goodresults.extend(results)
                prediction_scores.extend(scores)
            elif (model_key := (project.ls.id, label)) in self.segmentors:
                results, scores = await self.segmentors[model_key].predict_results(
                    features, label, cfg, project, task, context
                )
                goodresults.extend(results)
                prediction_scores.extend(scores)
            else:
                logging.warning(f"Skipping {label} of {project} because it has no classifier or segmentor")
        return {
            "result": goodresults,
            "model_version": self.version,
            "score": np.mean(prediction_scores) if len(prediction_scores) > 0 else 0,
        }

    @torch.no_grad()
    async def predict(self, project: LSProject, tasks: list[dict], context: dict | None = None):
        if len(tasks) >= 1:
            return {"results": [await self.predict_single(project, lstask, context) for lstask in tasks]}
        return {}

    async def train_classifier(self, label, cfg, tasks):
        assert len(cfg["inputs"]) == 1, "Choices with multiple inputs not supported yet"
        assert cfg["inputs"][0]["type"] == "Image", "Only image inputs supported for now"

        class_names = cfg["labels"]
        class_labels = self.get_classification_labels(tasks, label)
        X = self._clf_features["image"]
        y = np.array([class_names.index(class_labels[task_id.item()]) for task_id in self.task_ids["image"]])
        self.classifiers[label] = (
            lgb.train(
                {
                    "objective": "multiclass",
                    "min_data_in_leaf": min(y.shape[0] // len(class_names), 20),
                    "num_class": len(class_names),
                    "verbosity": -1,
                    # "num_boosting_round": 10,
                },
                lgb.Dataset(X.numpy(), label=y),
            ),
            class_names,
        )

    async def train(
        self,
        project: LSProject,
        tasks: list[int],
        mode: Literal["retrain", "continue"] = "retrain",
    ):
        logging.info(f"Training for {len(tasks)=}")

        # Add the new features to the datasets for each label
        for label, cfg in project.ls.parsed_label_config.items():
            for k in [inp["value"] for inp in cfg["inputs"] if inp["type"] == "Image"]:
                if k not in self._feature_buffer:
                    # self._dense_features[k], self._clf_features[k], self.task_ids[k] = self.get_features(tasks, k)
                    self._feature_buffer[k] = self.get_features(tasks, k)
            if cfg["type"] == "Choices":
                assert cfg["inputs"][0]["value"] == "image" and len(cfg["inputs"]) == 1, "Only image inputs supported"
                await self.train_classifier(label, cfg, tasks)
            elif cfg["type"] in ["BrushLabels"]:
                logging.info(f"Assuming region labels for {label}, training segmentor with {len(tasks)=}")
                if self.cfg.segmentation_mode == "multiclass":
                    if (project.ls.id, label) not in self.segmentors:
                        self.segmentors[(project.ls.id, label)] = await LGBMCSegmentor(self.cfg).update_model(
                            self._feature_buffer, label, cfg, tasks
                        )
                    else:
                        if mode == "retrain":
                            m = self.segmentors[(project.ls.id, label)]
                            m.model, m.X, m.y = None, None, None
                        await self.segmentors[(project.ls.id, label)].update_model(
                            self._feature_buffer, label, cfg, tasks
                        )
                elif self.cfg.segmentation_mode == "multilabel":
                    self.segmentors[(project.ls.id, label)] = await LGBMLSegmentor(self.cfg).update_model(
                        self._feature_buffer, label, cfg, tasks
                    )
                else:
                    raise ValueError(f"{self.cfg.segmentation_mode=} not supported")
            elif cfg["type"] in ["Brush"]:
                logging.debug(f"Skipping {cfg['type']} during training because its a tool, not a label")
            else:
                logging.error(f"{cfg['type']} not supported")
        self.version += 1

        # Release the features cache
        self._feature_buffer = {}

        # Delete all old predictions using the API if the model was retrained
        if mode == "retrain":
            pred_ids = await project.delele_all_predictions()
            logging.info(f"Deleting all predictions for {project}: {pred_ids=}")

        return {}

    def get_classification_labels(self, tasks: list[dict], label: str):
        tasks_with_label: dict[int, str] = {}
        for task in tasks:
            for result in task["annotations"][0]["result"]:
                if result["from_name"] == label:
                    tasks_with_label[task["id"]] = result["value"]["choices"][0]
        return tasks_with_label

    async def get_version(self):
        return {"model_version": f"{self.version}"}

    async def get_versions(self):
        return {"versions": [f"{self.version}"]}

    def _build_blurrer(self, like_tensor: torch.Tensor, kernel_size: int = 15, sigma: float = 2.5) -> nn.Module:
        kernel = _get_gaussian_kernel2d(
            (kernel_size, kernel_size),
            [sigma, sigma],
            like_tensor.dtype,
            like_tensor.device,
        ).repeat(like_tensor.shape[1], 1, 1, 1)
        blurrer = nn.Conv2d(
            like_tensor.shape[1],
            like_tensor.shape[1],
            kernel_size=kernel_size,
            stride=1,
            padding=(kernel_size - 1) // 2,
            padding_mode="replicate",
            groups=like_tensor.shape[1],
            bias=False,
        )
        blurrer.weight.data = kernel
        return blurrer

    @torch.inference_mode()
    def get_features_batch(self, torch_inputs: list[torch.Tensor]):
        """
        All torch_inputs should have the same shape like.
        """
        input_batch = torch.stack(torch_inputs, dim=0).to(self.torch_modules.get_device())

        feature_pyramid = self.torch_modules["encoder"].forward(input_batch)
        squeezed_map = self.torch_modules["squeezer"].forward(feature_pyramid)
        squeezed = self.flatten(self.pooling(squeezed_map))

        # Interpolate all feature maps from the feature_pyramid to the size of the pixel features
        pixel_features = [self.torch_modules["decoder"].forward(feature_pyramid)]
        # Pixel features shape: (batch_size, channels, height, width)

        interpolated = [
            F.resize(
                feature_map,
                size=pixel_features[0].shape[-2:],
                interpolation=F.InterpolationMode.NEAREST,
            )
            for feature_map in feature_pyramid[1:2]
        ]
        pixel_features.extend(interpolated)

        # Neighbourhood features enable the decision tree to also use the neighbourhood of a pixel
        if self.cfg.neighbourhood_features == "avgpool":
            pixel_features.append(
                nnF.avg_pool2d(
                    nnF.pad(pixel_features[0], (3, 3, 3, 3), mode="reflect"),
                    kernel_size=7,
                    stride=1,
                )
            )
        elif self.cfg.neighbourhood_features == "blur":
            blurrer = self._build_blurrer(pixel_features[0])
            pixel_features.append(blurrer(pixel_features[0]))
        elif self.cfg.neighbourhood_features == "none":
            logging.debug(f"Skipping neighbourhood features, because {self.cfg.neighbourhood_features=}")
        else:
            raise ValueError(f"{self.cfg.neighbourhood_features=} not supported")
        return torch.cat(pixel_features, dim=1).cpu(), squeezed.cpu()

    @torch.inference_mode()
    def get_features(self, tasks: dict, for_image: str):
        """
        For a batch of tasks, returns a tuple with

        - Dense features (num_pixels, #Channels): from the output of the decoder concatenated with the feature maps
        - Image features: (num_images, #Features): avg pooled from the output of the squeezer
        """
        input_images = [download_image(task["data"][for_image], ls_client=self.ls_client) for task in tasks]
        dimensions = [image.shape[:2] for image in input_images]
        torch_inputs = [F.to_tensor(input_image) for input_image in input_images]
        torch_inputs = [
            x["image"] for x in UnifySizes(divisable_by=32, enforce_order=True)([{"image": i} for i in torch_inputs])
        ]

        # Build batches, for now just process one image at a time
        batch_features = [self.get_features_batch([x]) for x in torch_inputs]

        return (
            [x[0] for x in batch_features],
            [x[1] for x in batch_features],
            torch.LongTensor([task["id"] for task in tasks]),
            dimensions,
        )

    async def get_instruction(self, project: Project, previous_instruction: str) -> tuple:
        def f_key(instruction):
            return "iframe" in instruction

        def f_instr(project, instruction):
            return self.env.get_template("lgbmodel_expertinstruction.jinja2").render({"modelurl": self.prefix})

        return f_key, f_instr



import logging
from typing import List, Tuple
from types import SimpleNamespace

import torch
import torch.nn as nn

from .SharedBlock import SharedBlock

try:
    from mmcv.cnn import ConvModule
    from mmdet.models.necks.fpn import FPN
except ImportError:
    ConvModule, FPN = None, None


class FCOSDecoder(SharedBlock):
    """
    FPN which converts encoder features into the features that the mmDetection task expects.
    """

    class Config(SharedBlock.Config):
        module_name: str = "fcosfpn"
        intermediate_channels: int = 128
        start_level: int = 1
        num_outputs: int = 5
        reg_convs: int = 4
        clf_convs: int = 4
        dcn_on_last_conv: bool = False

    def __init__(self, args: Config, enc_out_channels: List[int], enc_strides: List[int]) -> None:
        """
        Expects the full encoder channels (including channels of the input) and the full encoder strides.

        It filters them afterwards according to the config.
        """
        super().__init__(args)
        self.args: FCOSDecoder.Config
        # The first component describes the input and can be omitted here
        self.full_enc_channels, self.full_enc_strides = enc_out_channels, enc_strides
        self.enc_out_channels = enc_out_channels[1:]
        self.strides_after_start = enc_strides[self.args.start_level + 1 :]

        self.fpn: FPN = FPN(
            in_channels=self.enc_out_channels,  # example: [256, 512, 1024, 2048],
            out_channels=args.intermediate_channels,
            start_level=self.args.start_level,
            add_extra_convs="on_output",  # use P5
            num_outs=self.args.num_outputs,
            # norm_cfg=dict(type='BN'),
            # act_cfg=dict(type='GELU'),
            relu_before_extra_convs=True,
        )
        # self.reg_fpn: FPN = FPN(
        #     in_channels=self.enc_out_channels,  # example: [256, 512, 1024, 2048],
        #     out_channels=args.intermediate_channels,
        #     start_level=self.args.start_level,
        #     add_extra_convs='on_output',  # use P5
        #     num_outs=self.args.num_outputs,
        #     act_cfg=dict(type='ReLU'),
        #     relu_before_extra_convs=True)
        self.clf_convs = self.build_convstack(args.clf_convs, args.dcn_on_last_conv)
        self.reg_convs = self.build_convstack(args.reg_convs, args.dcn_on_last_conv)
        self.make_mtl_compatible()

    def get_strides(self) -> List[int]:
        # return e.g. [8, 16, 32, 64, 128]
        return self.strides_after_start + [
            self.strides_after_start[-1] * (i + 1) * 2
            for i in range(self.args.num_outputs - len(self.strides_after_start))
        ]

    def build_convstack(self, stackdepth: int, dcn_on_last_conv: bool) -> nn.Sequential:
        """Initialize bbox regression conv layers of the head."""
        res = nn.ModuleList()
        for i in range(stackdepth):
            # chn = self.in_channels if i == 0 else self.args.intermediate_channels
            if dcn_on_last_conv and i == stackdepth - 1:
                conv_cfg = dict(type="DCNv2")
            else:
                conv_cfg = None
            res.append(
                ConvModule(
                    self.args.intermediate_channels,
                    self.args.intermediate_channels,
                    3,
                    stride=1,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=dict(type="GN", num_groups=32, requires_grad=True),
                    act_cfg=dict(type="GELU"),
                    bias=True,
                )
            )
        return nn.Sequential(*res)

    def forward(self, feature_pyramid: List[torch.Tensor]) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        # Skip the first element, which is the input
        fpn_features = self.fpn(feature_pyramid[1:])
        # fpn_reg_features = self.reg_fpn(feature_pyramid[1:])
        clf_features = [self.clf_convs(x) for x in fpn_features]
        reg_features = [self.reg_convs(x) for x in fpn_features]

        return clf_features, reg_features

    def get_example_input(self):
        # using self.full_enc_channels, self.full_enc_strides
        example_input = [
            torch.rand(1, c, 224 // s, 224 // s).to(self.torch_device)
            for c, s in zip(self.full_enc_channels, self.full_enc_strides)
        ]
        return example_input



import logging
import os
from pathlib import Path
from typing import Optional, Literal
import torch
import torch.nn as nn
from mmm.BaseModel import BaseModel
from torchinfo import ModelStatistics, summary
from torchinfo.layer_info import LayerInfo


class MTLModule(nn.Module):
    """
    Baseclass for all tasks and all shared blocks.
    """

    class Config(BaseModel):
        module_name: str

    def __init__(self, args: Config) -> None:
        super().__init__()
        self.args = args
        self._epoch = -1
        self._prefix = ""
        self.torch_device: str = "cpu"
        self.training_state: Literal["frozen", "trainable"] = "trainable"
        assert True not in [
            s in self.args.module_name for s in [".", "/"]
        ], f"Do not use wandb special characters in name {self.args.module_name}"

    def set_device(self, device: Literal["cpu", "cuda"]):
        """
        Replacement for torch's to method which also tracks the device.
        """
        if device == "cuda":
            # For distributed training, the user script needs to call torch.cuda.set_device
            self = self.cuda()
        elif device == "cpu":
            self = self.cpu()
        else:
            raise ValueError(f"Unknown torch device {device} in MTLModule.set_device")
        self.torch_device = device
        return self

    @torch.no_grad()
    def reinit_parameters(self) -> None:
        """
        Generic resetting of torch.nn modules.
        """

        # Child modules
        def init_weights(m):
            if hasattr(m, "reset_parameters"):
                print(f"Resetting {m}")
                m.reset_parameters()

        self.apply(init_weights)

    def get_name(self) -> str:
        """The name that the task was given in its config."""
        return self.args.module_name

    def load_checkpoint(self, folder_path: Path):
        checkpoint = torch.load(folder_path / "module.ckpt", map_location=torch.device(self.torch_device))
        try:
            self.load_state_dict(checkpoint["model_state"])
        except RuntimeError as e:
            logging.error(f"Could not load checkpoint {folder_path} due to {e}. Retrying without strict mode")
            self.load_state_dict(checkpoint["model_state"], strict=False)

    def save_checkpoint(self, folder_path: Path):
        if not folder_path.exists():
            folder_path.mkdir(parents=True)
        save_dict = {"model_state": self.state_dict()}
        torch.save(save_dict, folder_path / "module.ckpt")

    def prepare_epoch(self, epoch: int, prefix: str, training_mode: Optional[bool] = None):
        """
        Called in the beginning of an epoch. Does not need to be overwritten by shared building blocks.
        """
        self._epoch = epoch
        self._prefix = prefix
        if training_mode is not None:
            if training_mode:
                # trainable parameters of norms would still update even with requires_grad==False:
                self.train(self.training_state == "trainable")
            else:
                self.eval()

    def freeze_all_parameters(self, freeze: bool = True):
        for param in self.parameters():
            param.requires_grad = not freeze
        self.training_state = "frozen" if freeze else "trainable"

    def __repr_html__(self) -> str:
        """
        Returns a html string describing this MTLModule.
        """
        stats: ModelStatistics = summary(model=self, verbose=0, depth=25)
        return f"""
                <b>Device:</b>{self.torch_device}<br />

                <pre><code>{stats.__repr__()}</pre></code>
                <pre><code>{self}</pre></code>

                """

    def get_short_status(self) -> str:
        return "-"



from typing import Any, Callable, TypeVar

from torch.utils.data import Dataset
import torchvision.transforms as transforms

from mmm.mtl_types import AnnotatedImage, RGBImage, ImageCaption
from mmm.data_loading.MTLDataset import MTLDataset, SrcCaseType


class ImageCaptionDataset(MTLDataset):
    """
    Each case consists of an image and a list of captions.
    """

    @staticmethod
    def batch_collater() -> Callable:
        """
        For detection tasks the common batch format seems to be a list of cases.
        By default, collate_fn overwriters work with a list of cases.
        """
        # No inputs are stacked, detection tasks want lists!
        batch_type = TypeVar("batch_type", bound=list[dict[str, Any]])

        def f(x: batch_type) -> batch_type:
            return x

        return f

    def __init__(self, src_ds: Dataset[SrcCaseType], collate_fn: Callable | None = None, *args, **kwargs) -> None:
        if collate_fn is not None:
            collate_fn = transforms.Compose([collate_fn, self.batch_collater()])
        else:
            collate_fn = self.batch_collater()
        super().__init__(src_ds, ["image", "captions"], ["meta"], collate_fn=collate_fn, *args, **kwargs)

    def verify_case(self, d: SrcCaseType) -> None:
        self.assert_image_data_assumptions(d["image"])
        assert "captions" in d, "ImageCaptionDataset requires a 'captions' field"
        assert isinstance(d["captions"], list), "ImageCaptionDataset requires a list of captions"

    def st_case_viewer(self, case: SrcCaseType, i: int = -1) -> None:
        from mmm.logging.st_ext import stw

        stw(AnnotatedImage.from_untyped(case))

    def _compute_batchsize_from_batch(self, batch: SrcCaseType) -> int:
        return len(batch)

    def _visualize_batch_case(self, batch: SrcCaseType, i: int) -> None:
        from mmm.logging.st_ext import stw

        stw(
            AnnotatedImage.from_untyped(
                {
                    "image": batch[i]["image"],
                    "captions": batch[i]["captions"],
                }
            )
        )



from __future__ import annotations
import json
from typing import Any, List, Dict, Tuple, Optional, Literal
from typing_extensions import Annotated
import random
import logging

import numpy as np
import torch.nn as nn
from pydantic import Field

from mmm.logging.type_ext import StepMetricDict
from mmm.logging.wandb_ext import build_wandb_image

from .MTLTask import MTLTask
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.data_loading.RegressionDataset import RegressionDataset
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.neural.pooling import GlobalPooling, GlobalPoolingConfig
from mmm.neural import LossConfigs, MSELossConfig
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules

from mmm.utils import flatten_list_of_dicts
from sklearn.metrics import mean_absolute_error, mean_squared_error, max_error, r2_score


class RegressionTask(MTLTask):
    """
    Special case of a MTLTask which deals with regression directly from the backbone's output $$Z$$.

    It requires cohorts holding regression datasets.
    """

    class Config(MTLTask.Config):
        encoder_key: str = "encoder"
        squeezer_key: str = "squeezer"
        loss_fn: Annotated[LossConfigs, Field(discriminator="loss_type")] = MSELossConfig()
        dropout: float = 0.2
        metrics: Optional[List[Literal["mae", "mse", "rmse", "max_error", "r2_score"]]] = Field(
            default=None,
            description="If none, the task will decide which metrics make sense",
        )
        head: Literal["pretraining", "smart"] = "pretraining"

    def __init__(self, hidden_dim: int, args: Config, cohort: TrainValCohort[RegressionDataset]):
        super().__init__(args, cohort)
        self.args: RegressionTask.Config  # Make sure IDE knows about the task specific fields
        self.hidden_dim = hidden_dim

        if self.args.head == "pretraining":
            self.task_modules = self._create_pretraining_head()
        else:
            self.task_modules = self._create_smart_head()
        self.flatten = nn.Flatten(1)
        self.criterion: nn.Module = self.args.loss_fn.build_instance()

    def _create_pretraining_head(self):
        head = nn.Sequential(nn.Dropout(p=self.args.dropout), nn.ReLU(), nn.Linear(self.hidden_dim, 1))

        return nn.ModuleDict({"regression_head": head})

    def _create_smart_head(self) -> nn.ModuleDict:
        head = nn.Sequential(
            nn.Dropout(p=self.args.dropout),
            nn.ReLU(),
            nn.Linear(max(4, self.hidden_dim), max(4, self.hidden_dim // 2)),
            nn.Dropout(p=self.args.dropout),
            nn.ReLU(),
            nn.Linear(self.hidden_dim // 2, 1),
        )

        return nn.ModuleDict({"regression_head": head})

    def prepare_batch(self, batch: Dict[str, Any]) -> Any:
        batch["image"] = batch["image"].to(self.torch_device)
        batch["target"] = batch["target"].unsqueeze(-1).to(self.torch_device)
        return batch

    def forward(self, inputs, shared_blocks: Dict[str, SharedBlock]):
        x, _ = inputs
        pyr = shared_blocks[self.args.encoder_key](x)
        _, squeezed = shared_blocks["squeezer"](pyr)
        hidden_vector = self.flatten(squeezed)

        return self.task_modules["regression_head"](hidden_vector)

    def training_step(self, batch: Dict[str, Any], shared_blocks: SharedModules):
        x = batch["image"]
        y = batch["target"]

        # skip if batch is empty
        tensornums = [bool(t.numel()) for t in x]
        if not True in tensornums:
            logging.info("encountered batch without valid training examples, skipping batch")
            return None

        y_hat = shared_blocks.forward((x, None), self.forward)

        batch_loss = self.criterion(y_hat, y)

        step_results: StepMetricDict = {  # type: ignore (.numpy() does not correctly indicate numpy array)
            "targets": y.cpu().numpy(),
            "preds": y_hat.detach().cpu().numpy(),
        }
        self.add_step_result(batch_loss.item(), step_results)

        live_vis = self._visualize_preds(
            x.detach().cpu(),
            step_results,
            (batch["meta"] if "meta" in batch else [{} for _ in range(batch["image"].shape[0])]),
        )

        return batch_loss, live_vis

    def _visualize_preds(self, training_ims, step_metrics: Dict, metas: List[Dict]) -> Dict[str, Any]:
        vis_n = min(self._takeout_vis_budget(), training_ims.size(0))

        if vis_n <= 0:
            return {}

        preds = []
        for rand_index in random.sample(list(range(training_ims.size(0))), vis_n):
            metastr = json.dumps(metas[rand_index], default=lambda o: str(o))
            description = f"{np.min(training_ims[rand_index].numpy()):.3f}, {np.max(training_ims[rand_index].numpy()):.3f}; {training_ims[rand_index].size()}"
            preds.append(
                build_wandb_image(
                    training_ims[rand_index],
                    f"{description}\nTrue: {step_metrics['targets'][rand_index]}\nPred: {step_metrics['preds'][rand_index]}\n{metastr}",
                )
            )
        return {"preds": preds} if preds else {}

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        metrics = flatten_list_of_dicts(self._step_metrics)
        if self.args.metrics is None:
            selected_metrics = ["mae", "rmse", "r2_score", "max_error"]
        else:
            selected_metrics = self.args.metrics

        _, print_str = super().log_epoch_metrics()
        log_dict: Dict[str, Any] = {}

        if "mae" in selected_metrics:
            log_dict["mae"] = mean_absolute_error(metrics["targets"], metrics["preds"])

        if "mse" in selected_metrics:
            log_dict["mse"] = mean_squared_error(metrics["targets"], metrics["preds"])

        if "rmse" in selected_metrics:
            log_dict["rmse"] = np.sqrt(mean_squared_error(metrics["targets"], metrics["preds"]))

        if "r2_score" in selected_metrics:
            log_dict["r2_score"] = r2_score(metrics["targets"], metrics["preds"])

        if "max_error" in selected_metrics:
            log_dict["max_error"] = max_error(metrics["targets"], metrics["preds"])

        return log_dict, print_str



"""
Utilities for streamlit. Use `stw` to visualize objects.

Importing this module will automatically set our recommended settings.
"""

import inspect
from typing import Optional, List, Callable, Dict, Tuple, Any
import traceback
import os
from functools import partial
from tqdm import tqdm
import random
import torch
import numpy as np
import cv2

try:
    import streamlit as st
except ImportError:
    st = None

from torch.utils.data import Dataset, DataLoader
from mmm.data_loading.MTLDataset import MTLDataset, DatasetStyle
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.DataSplit import DataSplit
from mmm.typing_utils import get_colors
from mmm.settings import mtl_settings

if st is not None:
    st.set_page_config(layout="wide")


def blend_with_mask(
    im_channels_first: torch.Tensor,
    mask_no_channels: Optional[torch.Tensor],
    caption_suffix: str = "",
    classes: Optional[List[str]] = None,
    st_key="default_blend_with_masks_key",
    boxes: Optional[Tuple[torch.FloatTensor, torch.LongTensor]] = None,
) -> None:
    """
    Shows images like they are expected by our neural networks.

    Images:
    - 2D: [C, width, height]; dtype: float32
    - 3D: [C, width, height, slices]: dtype: float32

    Mask:
    - 2D: [width, height]; dtype: long
    - 3D: [width, height, slices]; dtype: long
    """
    import streamlit as st

    caption = f"Shape: {tuple(im_channels_first.shape)}, "
    caption += f"Min: {torch.min(im_channels_first):.2f}, Max: {torch.max(im_channels_first):.2f}"

    if len(im_channels_first.shape) == 4:
        # 3D!
        slider_key = f"{st_key}_sliceslider"

        if slider_key not in st.session_state:
            st.session_state[slider_key] = random.randint(0, im_channels_first.shape[-1])

        st.slider(
            label="Select slice of image",
            min_value=0,
            max_value=im_channels_first.shape[-1] - 1,
            key=slider_key,
        )
        im_channels_first = im_channels_first[..., st.session_state[slider_key]]

        if mask_no_channels is not None:
            mask_no_channels = mask_no_channels[..., st.session_state[slider_key]]

    if torch.max(im_channels_first) > 1.0 or torch.min(im_channels_first) < 0.0:
        st.error("Pixel range is expected to be normalized into [0, 1]. Applying auto fix")
        im_channels_first = im_channels_first / 255.0
    im = im_channels_first.numpy().astype(np.float32)

    # Printing images happens channels last, while the library uses channels first
    im = np.moveaxis(im, 0, -1)

    # Writing annotations on top of the image without copying first can lead to errors
    im = im.copy()

    # For printing a color image is required. Scale the channels to 3
    if im.shape[-1] == 1:
        im = np.concatenate([im] * 3, axis=-1)

    if classes is None:
        classes_ind_name = [
            (mtl_settings.ignore_class_value, "Unlabeled"),
            (0, "BG"),
            (1, "FG"),
        ]
    else:
        classes_ind_name = [(mtl_settings.ignore_class_value, "Unlabeled")] + [(i, c) for i, c in enumerate(classes)]
    colors = get_colors(len(classes_ind_name))

    if mask_no_channels is not None:
        assert len(im_channels_first.shape) - len(mask_no_channels.shape) == 1, "Image needs channels, mask not"

        # Count the pixels of each class in the mask
        class_indices, class_counts = torch.unique(mask_no_channels, return_counts=True)
        for i, c in classes_ind_name:
            if i in class_indices:
                num_pixels = class_counts[class_indices == i][0]
            else:
                num_pixels = 0

            if num_pixels > 0:
                colorrgb = f"{int(colors[i][0]*255)}, {int(colors[i][1]*255)}, {int(colors[i][2]*255)}"
                col1, col2 = st.columns(2)
                with col1:
                    st.checkbox(
                        c,
                        key=f"{st_key}_{c}",
                        value=(len(classes_ind_name) == 1) or (i >= 0),
                    )
                with col2:
                    st.markdown(
                        f"<span style='color: rgb({colorrgb})'>{num_pixels} pixels</span>",
                        unsafe_allow_html=True,
                    )

            # Display the color of the class using markdown
            # st.markdown(f"<span style='color: rgb({colorrgb})'>asdf</span>", unsafe_allow_html=True)

        mask_no_channels = mask_no_channels.numpy()
        # color_overlay = np.zeros((mask_no_channels.shape[0], mask_no_channels.shape[1], 3))
        color_overlay = im.copy()
        for class_i, class_name in classes_ind_name:
            if f"{st_key}_{class_name}" in st.session_state and st.session_state[f"{st_key}_{class_name}"]:
                color_overlay[mask_no_channels == class_i] = np.asarray(colors[class_i], dtype=np.float32)

        # For each class, a binary mask is added in a distinct color
        # masks = []
        # for class_i, class_name in enumerate(classes):
        #     if st.session_state[f"{st_key}_{class_name}"]:
        #         binstr = f"{class_i:03b}"
        #         st.write(f"{class_name}, {binstr}")
        #         cc = []
        #         for c in binstr:
        #             if c == "1":
        #                 m_c = np.zeros_like(mask_no_channels)
        #                 m_c[mask_no_channels == class_i] = 1
        #                 cc.append(m_c)
        #             else:
        #                 cc.append(np.zeros_like(mask_no_channels))
        #         m = np.stack(cc, axis=-1).astype(np.float32)  # type: ignore
        #         # st.image(m)
        #         masks.append(m)

        # mm: np.ndarray = sum(masks) if masks else np.zeros_like(im)      # type: ignore

        im = cv2.addWeighted(im, 0.7, color_overlay.astype(np.float32), 0.3, 0)

    # Detection boxes
    if boxes is not None:
        boxes_tensor, box_labels = boxes
        for i, box in enumerate(boxes_tensor.tolist()):
            box_label: int = int(box_labels[i].item())
            thickness = max(1, im.shape[0] // 500)
            cv2.rectangle(
                im,
                list(map(int, box[:2])),
                list(map(int, box[2:])),
                colors[box_label],
                thickness,
            )

            fontscale: float = (im.shape[1] + im.shape[2]) / 1000
            fontsize_px: int = int(fontscale * 40)
            box_width = box[2] - box[0]
            chars_fitting = int(box_width / fontsize_px) * 2
            box_text = classes[box_label][:chars_fitting]
            cv2.putText(
                im,
                text=box_text,
                org=(int(box[0]), int(box[1]) + fontsize_px),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                color=colors[box_label],
                fontScale=fontscale,
            )

            # cv2 putText to write the label of the box
            # cv2.putText(im, f"{box_label}", (box[0], box[1]),
            #             cv2.FONT_HERSHEY_SIMPLEX, thickness, colors[box_label], thickness=thickness)

    st.image(im, caption=f"{caption}\n{caption_suffix}", clamp=True)


def _cohort_explorer(cohort: TrainValCohort) -> None:
    # Streamlit would otherwise resample cross validation splits
    import streamlit as st

    cohort.prepare_epoch(epoch=0)

    if split_name_selection := st.sidebar.selectbox("TrainOrVal", ["Training", "Validation"]):
        split_name: str = split_name_selection
    else:
        split_name: str = "Training"
    train_val_index = 0 if split_name == "Training" else 1

    def dataset_explorer():
        ds = cohort.datasets[train_val_index]
        st.title(f"{split_name} dataset:")

        if ds.get_dataset_style() is DatasetStyle.MapStyle:
            dataset_len = len(ds)  # type: ignore
            case_index = st.slider("select case", max_value=dataset_len - 1)
            case = ds.get_untransformed_case(case_index)
            ds.st_case_viewer(case, case_index)

            if st.button(label="verify all cases"):

                class VerifyingDataset(Dataset):
                    def __init__(self, src):
                        self.src = src

                    def __len__(self):
                        return len(self.src)

                    def __getitem__(self, i: int):
                        try:
                            case = self.src[i]
                            assert self.src.verify_case(case), f"case {case} could not be verified"
                            return case
                        except Exception as e:
                            print(i)
                            print(e)
                            st.write(i)
                            st.write(e)
                            traceback.print_exc()

                prog_bar = st.progress(0.0)
                st.write(f"Verifying with {len(os.sched_getaffinity(0))} workers")
                dl = DataLoader(
                    VerifyingDataset(ds),
                    shuffle=True,
                    num_workers=len(os.sched_getaffinity(0)),
                    batch_size=128,
                    collate_fn=lambda ls: ls,
                )
                for i, _ in enumerate(tqdm(dl)):
                    prog_bar.progress(i / (dataset_len // 128))
                st.balloons()
        else:
            with st.form("iterator demo"):
                max_items = int(st.number_input("Max iterations", step=1, value=10))
                display_every = int(st.number_input("Display every N case", step=1, value=2))
                submitted = st.form_submit_button("Reload iterator")

            # @st.cache(allow_output_mutation=True)
            # def cache_ds_iterator():
            #     return iter(ds)
            if submitted:
                for i, case in enumerate(ds):
                    ds.verify_case(case)
                    # case = next(iter(ds))
                    if i % display_every == 0:
                        ds.st_case_viewer(case, i)
                    if i > max_items:
                        break

    def batch_explorer():
        ds = cohort.datasets[train_val_index]
        test_batch = cohort.get_random_batch(DataSplit.from_index(train_val_index))
        ds.visualize_batch(test_batch)

    pages = {"Dataset explorer": dataset_explorer, "Batch explorer": batch_explorer}

    demo_name = st.sidebar.selectbox("Choose", list(pages.keys()))
    if demo_name is not None:
        pages[demo_name]()
    else:
        st.write("Select a page")


def side_by_side(img_1: torch.Tensor, img_2: torch.Tensor):
    """
    Shows two images side by side.
    Only implemented for 2D images [C,W,H], dtype: float32
    """
    import streamlit as st

    if torch.max(img_1) > 1.0 or torch.min(img_1) < 0.0:
        st.error("Pixel range is expected to be normalized into [0, 1]. Applying auto fix to first image")
    if torch.max(img_2) > 1.0 or torch.min(img_2) < 0.0:
        st.error("Pixel range is expected to be normalized into [0, 1]. Applying auto fix to second image")

    img_1: np.ndarray = img_1.numpy().astype(np.float32)
    img_2: np.ndarray = img_2.numpy().astype(np.float32)

    img_1 = np.moveaxis(img_1, 0, -1)
    img_2 = np.moveaxis(img_2, 0, -1)
    divider = np.zeros(img_1.shape)[:, :15, :]
    # Concatenating two images side by side

    img = np.concatenate((img_1, divider, img_2), axis=1)

    st.image(
        img,
        caption=f"two views of the same image. each of shape {img_1.shape}",
        clamp=True,
    )


def stw(obj: Any, st_prefix: str = "") -> None:
    """
    Takes an object and tries to visualize it.

    If the package did not define a special rule for that object we call streamlit's write.
    """
    if hasattr(obj, "_st_repr_"):
        # If the object's st_repr has a keyword parameter st_prefix, use the global counter
        if "st_prefix" in inspect.signature(obj._st_repr_).parameters:
            st_repr = obj._st_repr_(st_prefix=st_prefix)
        else:
            st_repr = obj._st_repr_()
        if st_repr is not None:
            st.markdown(
                st_repr,
                unsafe_allow_html=True,
            )
        return None
    return st.write(obj)


def multi_cohort_explorer(cohorts: Dict[str, Callable[[], TrainValCohort]]):
    # please use chromium-based browsers for correct image display
    # torch.manual_seed(0)
    # np.random.seed(0)
    # random.seed(0)
    import streamlit as st

    cohort_name = st.sidebar.selectbox("Choose cohort", list(cohorts.keys()))

    if cohort_name:
        stw(cohorts[cohort_name]())
    else:
        st.write(f"Select a Cohort")
